#!/usr/bin/python3

""" raster_summary.py
	Provides a summary of the raster metadata and plot the image.
	User ensures the raster file path is known and pass into this function.
	Note the naming scheme, '.tif' is the common type when raster is originally created.

	https://rasterio.readthedocs.io/en/latest/topics/reading.html

	This function opens the file for reading then calls the close() method to end
	the process.  If this function gets called many times, this could slow down
	the process.  Remove the close() at the end or redesign the code to avoid
	process slowdown.

	Libraries:
		OS
		Pprint
		Rasterio

	API Calls:
		os.path.join()
		rasterio.open()
		pprint()
		show()

        External Dependency:
		None

"""

import os
from pprint import pprint
import rasterio
from rasterio.plot import show


def raster_summary(raster_dir: str, raster_file: str):
    """Reads a raster file and outputs the metadata and image.

    Args:
            raster_dir:	dir path location of the raster file
            raster_file:	raster file name
    Returns:
            None
    """

    # for single file operation, use line by line commands
    # open file for rasterio
    fp = os.path.join(raster_dir, raster_file)

    # Open the file:
    raster = rasterio.open(fp)

    # from rasterio doc: attributes
    print(f"Raster shape:\t\t {raster.shape}")
    print(f"Raster band count:\t {raster.count}")
    print(f"Raster data types:\t {raster.dtypes} ")
    print(f"Raster valid data mask:\t {raster.nodatavals}")
    print(f"Raster not valid mask:\t {raster.nodata}")

    print(f"Raster metadata:\n")
    pprint(raster.meta)

    # the plot dimensions show the longitude, x, and lattitude, y
    show(raster)

    raster.close()
    return


# end raster_summary



import rasterio


def resolution(input_tiff):
    """
    Calculate the resolution of the given TIF file.

    :param input_tiff: Path to TIF file.
    :type input_tiff: str

    :returns: resolution as a (x, y) tuple
    """

    tiff = rasterio.open(input_tiff)
    # get dimensions, in units
    # 1 degree latitude is approx 111 kms
    # 1 degree longitude is approx 111 kms

    # longitude
    width = (
        (tiff.bounds.right - tiff.bounds.left) * 111 * 1000
    )  # converting kms to meters
    # latitude
    height = (tiff.bounds.top - tiff.bounds.bottom) * 111 * 1000

    # get dimentions in pixels
    px_width = tiff.width
    px_height = tiff.height
    print("Width {} , height {} in pixels".format(px_width, px_height))

    # meters in one pixel
    w = width / px_width
    h = height / px_height

    return (w, h)



import re
import warnings
from itertools import groupby
from typing import List, Set, Union

import pandas as pd
import spacy


def doc2df(doc: spacy.tokens.Doc, regex_pmid: str = r"PMID: (\d+)") -> pd.DataFrame:
    """
    Get a dataframe from SpaCy `Doc` objects.

    :param doc: A single spaCy `Doc` object with annotations (entities)
    :type doc: spacy.tokens.Doc
    :param regex_pmid: Raw string with regex to get the PMID numbers from text (defaults to `r"PMID: (\\d+)"`
    :type regex_pmid: str
    :return: Pandas dataframe with columns [`pmid`,`text`,`start`,`stop`,`entity`,`entity_text`]
    """

    match = re.search(regex_pmid, doc.text)
    if match:
        pmid = match.group(1)
    else:
        raise ValueError("Could not find PMID.")
    if len(doc.ents) <= 0:
        annotations = [(None, None, None, None)]
        warnings.warn(f"No annotations found on document with PMID `{pmid}`.")
    else:
        annotations = [(e.start_char, e.end_char, e.label_, e.text) for e in doc.ents]
    start, stop, entity, ent_text = zip(*annotations)
    df = pd.DataFrame(
        {
            "pmid": [pmid] * len(annotations),
            "text": [doc.text] * len(annotations),
            "start": start,
            "stop": stop,
            "entity": entity,
            "entity_text": ent_text,
        }
    )
    return df


def doc2df_wide(
    doc: spacy.tokens.Doc, regex_pmid: str = r"PMID: (\d+)"
) -> pd.DataFrame:
    """
    Get a wide dataframe from SpaCy `Doc` objects.

    :param doc: A single spaCy `Doc` object with annotations (entities)
    :type doc: spacy.tokens.Doc
    :param regex_pmid: Raw string with regex to get the PMID numbers from text (defaults to `r"PMID: (\\d+)"`
    :type regex_pmid: str
    :return: Pandas dataframe with one column for each {`start`,`stop`,`entity_text`} in entity
    """

    df = doc2df(doc=doc, regex_pmid=regex_pmid).reset_index(drop=True)
    # catch edge-case where there are no entities predicted in document
    if df["entity"].isna().any():
        return df[["pmid", "text"]].set_index("pmid")

    # pivot table (aggregate duplicate pmid entries using lists)
    df = df.pivot_table(
        index=["pmid", "text"],
        columns="entity",
        values=["start", "stop", "entity_text"],
        aggfunc=list,
    ).swaplevel(axis="columns")
    df.columns = df.columns.to_flat_index()

    return df.sort_index(axis=1).reset_index("text")


def df2jsonl(
    df: pd.DataFrame,
    path: str = None,
    pmid: str = "pmid",
    text: str = "text",
    start: str = "start",
    stop: str = "stop",
    entity: str = "entity",
    entity_text: str = "entity_text",
) -> Union[str, None]:
    """
    Get the output format JSONL from a pandas dataframe

    :param df: Dataframe with entity information
    :type df: pd.DataFrame
    :param path: Destination file path (if None will return a string with equivalent JSONL
    :type path: str
    :param pmid: Column name for PMID column
    :type pmid: str
    :param text: Column name for text column
    :type text: str
    :param start: Column name for start column
    :type start: str
    :param stop: Column name for stop column
    :type stop: str
    :param entity: Column name for entity column
    :type entity: str
    :param entity_text: Column name for entity_text column
    :type entity_text: str
    :return: None if path is provided (file is written to disk), JSONL string if path is None
    Output JSONL file has the shape `{id:pmid, text: <abstract>, predictions: [{start: x, end: y, entity:x}, ...]}`
    """
    _df = df.copy()
    _df = _df.reset_index(drop=True)
    _df["predictions"] = pd.Series(
        zip(_df[start], _df[stop], _df[entity], _df[entity_text])
    ).apply(lambda x: dict(zip(["start", "end", "entity", "entity_text"], x)))
    _df = _df.groupby([pmid, text])["predictions"].apply(list).reset_index()
    return _df.to_json(path, orient="records", lines=True)


def list2iob(tags: List[str]) -> List[str]:
    """
    Go from a list of strings to IOB list of tags.

    :param tags: List of tags
    :type tags: List[str]

    :return: List of IOB tags
    """

    def _add_iob_prefix(idx: int, tag: str) -> str:
        """
        Add IOB prefix to sequence of repeated tags.
        """
        return tag if tag == "O" else "B-" + tag if idx == 0 else "I-" + tag

    return [
        _add_iob_prefix(i, grp) for _, grp in groupby(tags) for i, grp in enumerate(grp)
    ]


def adjust_tags(
    tags: List[str], tag_set: Set[str] = None, length: int = 512
) -> List[List[str]]:
    """
    Keep only tags that were used in training and get all list of tokens to the same length.

    :param tags: List of tags
    :type tags: List[str]
    :param tag_set: Set of tags to keep
    :type tag_set: Set[str]
    :param length: Length of tokens to keep

    :return: List of tags
    """

    def _pad_and_truncate(tags_doc: List[str], seq_len: int) -> List[List[str]]:
        """Get all the list of tags to the correct length."""
        tags_doc = tags_doc + ["O"] * (seq_len - len(tags_doc))  # pad
        tags_doc = tags_doc[:512]  # truncate
        return tags_doc

    if tag_set is None:
        tag_set = set(tags)
    else:
        # put tag set in IOB format
        tag_set = sorted(
            [tag.split("-")[-1] for tag in tag_set] * 2
        )  # go to list and duplicate
        tag_set = list2iob(tag_set)  # make IOB set

    # keep tag if it's in the tag set
    tags = [tag if tag in tag_set else "O" for tag in tags]
    return _pad_and_truncate(tags, seq_len=length)



import numpy as np

def labels_to_class_weights(labels, nc=None):
    """ 
    Get class weights (inverse frequency) from training labels

    :param labels: Target classes, list or numpy array if n-dimensional array then
                   classes need to be in the 0-dim.
                   examples: 
                   
                   Labels can be list

                   >>> labels =  [1,2,3,4,5] 

                   or n-dim array

                   >>> labels = [[[class1,x1,y1,x2,y2],
                                  [class2,x1,y1,x2,y2],
                                 ...
                                ]]


    :type labels: list, numpy.ndarray, 
    :param nc: (Optional) Minimum count of class occurance.
    :type nc: int

    return(list) = list of weights

    :example:

    >>> labels = [0,1,2,3,1,1,2,3,4,5,5,5]
    >>> labels_to_class_weights(labels)
    [0.27272727 0.09090909 0.13636364 0.13636364 0.27272727 0.09090909]

    >>> labels = np.array([[[0,0.2,0.2,0.3,0.5],[1,0.2,0.2,0.3,0.4],[1,0.2,0.3,0.4,0.5]]])
    >>> labels_to_class_weights(labels)
    [0.66666667 0.33333333]

    >>> labels = [[1],[2],[3],[4]]
    >>> labels_to_class_weights(labels)
    [0.2 0.2 0.2 0.2 0.2]
    
    """

    assert labels[0] is not None, "no labels passed"

    if type(labels) == list:
        labels = np.array(labels)

    if len(labels.shape) > 1:
        labels = np.concatenate(labels, 0)  # labels.shape = (866643, 5) for COCO
        
        # after concat if till not vector for case [[1],[2],[3],[4]]
        if len(labels.shape) > 1:
            labels = labels[:, 0].astype(np.int)  # labels = [class xywh]
    
    weights = np.bincount(labels, minlength=nc)  # occurrences per class
    # Prepend gridpoint count (for uCE training)
    # gpi = ((320 / 32 * np.array([1, 2, 4])) ** 2 * 3).sum()  # gridpoints per image
    # weights = np.hstack([gpi * len(labels)  - weights.sum() * 9, weights * 9]) ** 0.5  # prepend gridpoints to start

    weights[weights == 0] = 1  # replace empty bins with 1
    weights = 1 / weights  # number of targets per class
    weights /= weights.sum()  # normalize

    return weights



# Adapted from: https://github.com/rwightman/pytorch-image-models/blob/bd5694667625d510e3aa89e95397de97d0e7efe9/benchmark.py

import time
import os
from contextlib import suppress
from functools import partial
import torch
import torch.nn as nn
from .utils import Params

amp = False
try:
    if getattr(torch.cuda.amp, "autocast") is not None:
        amp = True
except AttributeError:
    pass
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def timestamp(sync: bool = False):
    """
    returns time.perf_counter
    """
    return time.perf_counter()


def cuda_timestamp(sync: bool = False, device=None):
    """
    synchronizes cuda device if sync is true

    :param sync: boolean value deciding whether to sync
    :type sync: boolean
    :param device: cuda device
    :type device: torch.cuda.device
    """
    if sync:
        torch.cuda.synchronize(device=device)
    return time.perf_counter()


def count_params(model: nn.Module):
    """
    returns the number of parameters of the model

    :param model: neural network model
    :type model: nn.Module
    """
    return sum([m.numel() for m in model.parameters()])


def resolve_precision(precision: str):
    """
    Resolves the precision of the data type

    :param precision: precision data type passed in
    :type precision: str
    """
    assert precision in ("float16", "bfloat16", "float32")
    use_amp = False
    model_dtype = torch.float32
    data_dtype = torch.float32
    if amp:
        use_amp = True
    elif precision == "float16":
        model_dtype = torch.float16
        data_dtype = torch.float16
    elif precision == "bfloat16":
        model_dtype = torch.bfloat16
        data_dtype = torch.bfloat16
    return use_amp, model_dtype, data_dtype


class BenchmarkRunner:
    """
    Base class which can be extended to inference or traning benchmark.

    :param model_dir: path to folder containing `params.json` file
    :type model_dir: str
    :param num_warm_iter: number of iterations as warmup
    :type num_warm_iter: int
    :param num_bench_iter: number of iterations to benchmark
    :type num_bench_iter: int
    :param precision: precision to use for benchmarking ["float16", "float32", "bfloat16"]
    :type precision: str

    """

    def __init__(
        self,
        model_dir: str,
        num_warm_iter: int = 10,
        num_bench_iter: int = 50,
        precision: str = "float16",
    ):
        """
        :param model_dir: directory from where the model is loaded
        :type model_dir: str
        :param num_warm_iter: warming steps for training
        :type num_warm_iter: int
        :param num_bench_iter: benchmark iterations
        :type num_bench_iter: int
        :param precision: precision value for threshold
        :type precision: str
        """

        json_path = os.path.join(model_dir, "params.json")
        assert os.path.isfile(
            json_path
        ), "No json configuration file found at {}".format(json_path)
        params = Params(json_path)
        self.use_amp, self.model_dtype, self.data_dtype = resolve_precision(precision)
        self.amp_autocast = torch.cuda.amp.autocast if self.use_amp else suppress
        print(f"Loading model from path {params.save_model_path}")
        self.model = torch.load(params.save_model_path).to(
            device=device, dtype=self.model_dtype
        )
        self.param_count = count_params(self.model)
        self.num_classes = params.num_classes
        print("Model created, param count: %d" % (self.param_count))
        self.data_h, self.data_w, self.data_d = (
            params.height,
            params.width,
            params.input_channels,
        )
        self.input_size = (self.data_d, self.data_h, self.data_w)
        self.batch_size = params.batch_size
        self.num_warm_iter = num_warm_iter
        self.num_bench_iter = num_bench_iter
        self.log_freq = num_bench_iter // 5
        if torch.cuda.is_available():
            self.time_fn = partial(cuda_timestamp, device=device)
        else:
            self.time_fn = timestamp

    def _init_input(self):
        """
        make example inputs
        """
        # (NCHW)
        self.example_inputs = torch.randn(
            (self.batch_size,) + self.input_size, device=device, dtype=self.data_dtype
        )


class InferenceBenchmarkRunner(BenchmarkRunner):
    """
    Inference class extended from :class:`BenchmarkRunner`

    :param model_dir: path to folder containing `params.json` file
    :type model_dir: str
    :param precision: precision to use for benchmarking ["float16", "float32", "bfloat16"]
    :type precision: str

    """

    def __init__(self, model_dir: str, precision: str = "float32"):
        """
        :param model_dir: directory where to load the model from
        :type model_dir: str
        :param precision: data type precision
        :type precision: str
        """
        super().__init__(model_dir=model_dir, precision=precision)
        self.model.eval()

    def run(self):
        """
        Run the neural network model
        """

        def _step():
            t_step_start = self.time_fn()
            with self.amp_autocast():
                output = self.model(self.example_inputs)
            t_step_end = self.time_fn(True)
            return t_step_end - t_step_start

        print(
            f"Running inference benchmark on the model for {self.num_bench_iter} steps w/ "
            f"input size {self.input_size} and batch size {self.batch_size}."
        )

        with torch.no_grad():
            self._init_input()

            for _ in range(self.num_warm_iter):
                _step()

            total_step = 0.0
            num_samples = 0
            t_run_start = self.time_fn()
            for i in range(self.num_bench_iter):
                delta_fwd = _step()
                total_step += delta_fwd
                num_samples += self.batch_size
                num_steps = i + 1
                if num_steps % self.log_freq == 0:
                    print(
                        f"Infer [{num_steps}/{self.num_bench_iter}]."
                        f" {num_samples / total_step:0.2f} samples/sec."
                        f" {1000 * total_step / num_steps:0.3f} ms/step."
                    )
            t_run_end = self.time_fn(True)
            t_run_elapsed = t_run_end - t_run_start

        results = dict(
            samples_per_sec=round(num_samples / t_run_elapsed, 2),
            step_time=round(1000 * total_step / self.num_bench_iter, 3),
            batch_size=self.batch_size,
            img_size=self.input_size[-1],
            param_count=round(self.param_count / 1e6, 2),
        )

        print(
            f"Inference benchmark of the model done. "
            f"{results['samples_per_sec']:.2f} samples/sec, {results['step_time']:.2f} ms/step"
        )
        return results



import re
from typing import List
import spacy
from spacy.tokens import Doc


class TextPreprocessor:
    """
    Pre process texts

    :param pos_tags_removal: list of PoS tags to remove
    :type pos_tags_removal: List[str]
    :param spacy_model: spaCy model to be used, default is en_core_web_sm
    :param remove_numbers: whether or not to remove numbers from the inputs
    :type remove_numbers: bool
    :param remove_special_chars: whether or not to remove the special characters
    :type remove_special_chars: bool
    :param remove_stopwords: whether or not to remove stopwords
    :type remove_stopwords: bool
    :param lemmatize: whether or not to lemmatize the inputs
    :type lemmatize: bool

    :Example:

    >>> model = TextPreprocessor.load_model()
    >>> preprocessor = TextPreprocessor(spacy_model=model)
    >>> sample = "Hello wROld! I hAvE $400 Dollars in my checking account and I am sleeping"
    >>> clean = preprocessor.preprocess_text(sample)
    >>> print(f"Sample: {sample} Cleaned: {clean}")

    Sample: Hello wROld! I hAvE $400 Dollars in my checking account and I am sleeping

    Cleaned: hello wrold dollar checking account sleep

    >>> my_list = ["my name is $Yudhiesh", "that house is far AWAY", "the placement @RT"]
    >>> cleaned_list = preprocessor.preprocess_text_lists(my_list)
    >>> print(f"Sample: {my_list} Cleaned: {cleaned_list}")

    Sample: ['my name is $Yudhiesh', 'that house is far AWAY', 'the placement @RT']

    Cleaned: ['yudhiesh', 'house far away', 'placement  rt']

    """

    def __init__(
        self,
        pos_tags_removal=None,
        spacy_model=None,
        remove_numbers: bool = True,
        remove_special_chars: bool = True,
        remove_stopwords: bool = True,
        lemmatize: bool = True,
    ) -> None:
        """
        :param pos_tags_removal: list of PoS tags to remove
        :type pos_tags_removal: List[str]
        :param spacy_model: spaCy model to be used, default is en_core_web_sm
        :param remove_numbers: whether or not to remove numbers from the inputs
        :type remove_numbers: bool
        :param remove_special_chars: whether or not to remove the special
        characters
        :type remove_special_chars: bool
        :param remove_stopwords: whether or not to remove stopwords
        :type remove_stopwords: bool
        :param lemmatize: whether or not to lemmatize the inputs
        :type lemmatize: bool

        :Example:

        >>> model = TextPreprocessor.load_model()
        >>> preprocessor = TextPreprocessor(spacy_model=model)
        >>> sample = "Hello wROld! I hAvE $400 Dollars in my checking account and I am sleeping"
        >>> clean = preprocessor.preprocess_text(sample)
        >>> print(f"Sample: {sample}\nCleaned: {clean}")
        Sample: Hello wROld! I hAvE $400 Dollars in my checking account and I am sleeping
        Cleaned: hello wrold dollar checking account sleep

        >>> my_list = ["my name is $Yudhiesh", "that house is far AWAY", "the placement @RT"]
        >>> cleaned_list = preprocessor.preprocess_text_lists(my_list)
        >>> print(f"Sample: {my_list}\nCleaned: {cleaned_list}")
        Sample: ['my name is $Yudhiesh', 'that house is far AWAY', 'the placement @RT']
        Cleaned: ['yudhiesh', 'house far away', 'placement  rt']
        """
        self.remove_numbers = remove_numbers
        self.remove_special_chars = remove_special_chars
        self.remove_stopwords = remove_stopwords
        self.lemmatize = lemmatize
        self.pos_tags_removal = pos_tags_removal

        if not spacy_model:
            self.model = spacy.load("en_core_web_sm")
        else:
            self.model = spacy_model

    @staticmethod
    def download_spacy_model(model="en_core_web_sm"):
        """
        Downloads a spaCy model to be used for preprocessing the text inputs

        """
        print(f"Downloading spaCy model {model}")
        spacy.cli.download(model)
        print("Completed downloading the model")

    @staticmethod
    def load_model(model="en_core_web_sm"):
        """
        Loads and returns a spaCy model

        """
        return spacy.load(model, disable=["ner", "parser"])

    def preprocess_text(self, text: str) -> str:
        """
        Preprocesses a single string

        :param text: text string to clean
        :return: str

        """
        doc = self.model(text)
        return self.__preprocess(doc)

    def preprocess_text_lists(self, texts: List[str]) -> List[str]:
        """
        Preprocesses a list of strings

        :param texts: List of texts to preprocess
        :return: List of strings

        """
        return [self.__preprocess(doc) for doc in self.model.pipe(texts)]

    def __preprocess(self, doc: Doc) -> str:

        # POS Tags Removal
        if self.pos_tags_removal:
            tokens = [token for token in doc if token.pos_ not in self.pos_tags_removal]
        else:
            tokens = doc

        # Remove numbers
        if self.remove_numbers:
            tokens = [
                token for token in tokens if not (token.like_num or token.is_currency)
            ]

        # Remove stopwords
        if self.remove_stopwords:
            tokens = [token for token in tokens if not token.is_stop]

        # Remove unwated tokens
        tokens = [
            token
            for token in tokens
            if not (
                token.is_punct or token.is_space or token.is_quote or token.is_bracket
            )
        ]

        # Remove empty tokens
        tokens = [token for token in tokens if token.text.strip() != ""]

        # Lemmatize
        if self.lemmatize:
            text = " ".join([token.lemma_ for token in tokens])
        else:
            text = " ".join([token.text for token in tokens])

        # Remove non-alphabetic chars

        if self.remove_special_chars:
            text = re.sub(r"[^a-zA-Z\']", " ", text)

        # Remove non-Unicode chars
        text = re.sub(r"[^\x00-\x7F]+", "", text)

        return text.lower()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import cv2
import albumentations as A
from albumentations.pytorch import ToTensorV2


def get_basic_train_transforms(height: int, width: int, means: list, stds: list):
    """
    Apply only basic training transformations such as Resize and Normalize.

    :param height: int specifying new height
    :param width: int specifying new width
    :param means: List of means for normalization
    :param stds: List of stds for normalization
    :returns: Albumentation compose transform object for training dataset
    """
    trn_transform = A.Compose(
        [
            A.Resize(height, width, cv2.INTER_NEAREST),
            A.Normalize(mean=means, std=stds),
            ToTensorV2(),
        ]
    )
    return trn_transform


def get_mild_train_transforms(height: int, width: int, means: list, stds: list):
    """
    Apply few mild training transformations such as Resize, horizontal and vertical, Gaussian Noise,
    Perspective Shift and Normalize.

    :param height: int specifying new height
    :param width: int specifying new width
    :param means: List of means for normalization
    :param stds: List of stds for normalization
    :returns: Albumentation compose transform object for training dataset
    """
    trn_transform = A.Compose(
        [
            A.Resize(height, width, cv2.INTER_NEAREST),
            A.HorizontalFlip(p=0.5),
            A.VerticalFlip(p=0.5),
            A.GaussNoise(p=0.2),
            A.Perspective(p=0.5),
            A.Normalize(mean=means, std=stds),
            ToTensorV2(),
        ]
    )
    return trn_transform


def get_val_transforms(height: int, width: int, means: list, stds: list):
    """
    Apply only basic transformation such as Resize and Normalize.
    :param height: int specifying new height
    :param width: int specifying new width
    :param means: List of means for normalization
    :param stds: List of stds for normalization
    :returns: Albumentation compose transform object for validation dataset
    """
    val_transform = A.Compose(
        [
            A.Resize(height, width, cv2.INTER_NEAREST),
            A.Normalize(mean=means, std=stds),
            ToTensorV2(),
        ]
    )
    return val_transform



import re


def noise_removal(input_text, noise_list=[]):
    """
    Remove noise from the input text.
    This function can be use to remove comman english words like "is, a , the".
    You can add your own noise words in the optional paramters "noise_list"
    default noise_list = ["is", "a", "this", "the" , "an"]

    :param input_text: Input text to remove noise from
    :type input_text: str
    :param noise_list: (Optional) List of words that you want to remove from the input text
    :type noise_list: list
    :returns: Clean text with no noise
    """

    assert isinstance(input_text, str), "input should be a string"
    if len(noise_list) == 0:
        noise_list = ["is", "a", "this", "the", "an"]

    words = input_text.split()
    noise_free_words = [word for word in words if word not in noise_list]
    noise_free_text = " ".join(noise_free_words)
    return noise_free_text


def remove_regex_pattern(input_text, regex_pattern):
    """
    Remove any regular expressions from the input text

    :param input_text: Input text to remove regex from
    :type input_text: str
    :param noise_list: string of regex that you want to remove from the input text
    :type noise_list: string
    :returns: Clean text with removed regex

    :example:

    >>> regex_pattern = "#[\w]*"
    >>> remove_regex_pattern("remove this #hastag" , regex_pattern)
    "remove this  hastag"
    """

    urls = re.finditer(regex_pattern, input_text)
    for i in urls:
        input_text = re.sub(i.group().strip(), "", input_text)

    return input_text


def remove_hashtags(text):
    """
    Removing hastags from the input text
    :param text: Input text to remove hastags from
    :type input_text: str
    :returns: output text without hastags

    :example:

    >>> remove_hashtags("I love #hashtags")
    "I love "
    """
    return remove_regex_pattern(text, r"(#\w+)")


def remove_url(text):
    """
    Removing urls from the input text
    :param text: Input text to remove urls from
    :type input_text: str
    :returns: text with standard words

    :example:

    >>> remove_urls('I like urls http://www.google.com')
    'I like urls '
    """
    return remove_regex_pattern(text, r"(#\w+)")


def standardize_words(input_text, lookup_dictionary):
    """
    Replace acronyms, hastags, colloquial slangs etc with standard words

    Text data often contains words or phrases which are not present in any standard lexical dictionaries.
    These pieces are not recognized by search engines and models.

    :param input_text: Input text to remove regex from
    :type input_text: str
    :param lookup_dictionary: Dictionary with slang as index and standard word as item or value
    :type lookup_dictionary: dict
    :returns: text with standard words

    :example:

    >>> lookup_dict = {'rt':'Retweet', 'dm':'direct message', "awsm" : "awesome", "luv" :"love", "..."}
    >>> standardize_words(rt I luv to dm, it's awsm)
    Retweet I love to direct message, it's awesome
    """
    words = input_text.split()
    new_words = []
    for word in words:
        if word.lower() in lookup_dictionary:
            word = lookup_dictionary[word.lower()]
        new_words.append(word)
        new_text = " ".join(new_words)
    return new_text



import numpy as np
from sklearn import metrics
import seaborn as sns
import matplotlib.pyplot as plt


def plot_cm(true, preds, classes: list, figsize: tuple = (8, 6)):
    """
    Plot unnormalized confusion matrix

    :param true: List of targets
    :param preds: List of predictions
    :param classes: List of classes
    :param figsize: Tuple specifying (height, width)
    :returns: matplotlib figure containing confusion matrix
    """
    cm = metrics.confusion_matrix(true, preds)
    fig = plt.figure(figsize=figsize)
    sns.heatmap(
        cm,
        xticklabels=classes,
        yticklabels=classes,
        annot=True,
        fmt="d",
        cmap="Blues",
        vmin=0.2,
    )
    plt.title("Confusion Matrix")
    plt.ylabel("True Class")
    plt.xlabel("Predicted Class")
    return fig


def unnormalize_image(img, means, stds):
    """
    Convert normalized image to get unnormalized image

    :param img: Tensor of shape (C, H, W)
    :param means: List of means used for normalization
    :param stds: List of stds used for normalization
    :returns: the unnormalized input tensor which can be used to display image
    """
    img = img.numpy()
    img = np.transpose(img, (1, 2, 0))
    # unnormalize
    img = img * np.array(stds) + np.array(means)
    return img


def plot_hist(history):
    """
    Plotting train acc, loss and val acc and loss stored in history dict.
    History dict contains keys = {train_acc, val_acc, train_loss, val_loss}
    Each key contains list of scores for every epoch.

    :param history: Dict
    :returns: plot the loss and acc plots for train and val
    """
    # summarize history for accuracy
    plt.plot(history["train_acc"])
    plt.plot(history["val_acc"])
    plt.title("model accuracy")
    plt.ylabel("accuracy")
    plt.xlabel("epoch")
    plt.legend(["train", "val"], loc="upper left")
    plt.show()
    # summarize history for loss
    plt.plot(history["train_loss"])
    plt.plot(history["val_loss"])
    plt.title("model loss")
    plt.ylabel("loss")
    plt.xlabel("epoch")
    plt.legend(["train", "val"], loc="upper left")
    plt.show()



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.layers import (
    Input,
    concatenate,
    MaxPooling2D,
    Conv2D,
    UpSampling2D,
)
from tensorflow.keras.models import Model
from tensorflow.keras.regularizers import l2
from tensorflow.keras import backend as K


def greenpixeldetection(pretrained_weights, optimizer, loss, input_size):
    """
    detect the green pixels in an image.

    Originally, there’s a library, PlantCV that does this using non neural network methods,
    since PlantCV only runs on yhe CPU, we developed this as an alternative method that could run on the GPU

    :param pretrained_weights: pretrained weight matrix
    :param input_size: image dimension tuple => (img_height, img_width, input_channel)
    :param optimizer: optimization strategy
    :param loss: loss function
    :type pretrained_weights: numpy ndarray
    :type input_size: tuple
    :type optimizer: optimizer object
    :type loss: loss object
    :return: compiled Keras model
    :rtype: Keras model object
    """

    inputs = Input(input_size)

    conv1_1 = Conv2D(16, (3, 3), padding="same")(inputs)
    pool1 = MaxPooling2D(2, 2)(conv1_1)

    conv3_1 = Conv2D(32, (5, 5), activation="relu", padding="same")(pool1)
    conv3_1 = Conv2D(32, (5, 5), activation="relu", padding="same")(conv3_1)

    up2_1 = UpSampling2D(size=2, interpolation="bilinear")(conv3_1)

    output = concatenate([up2_1, conv1_1], axis=3)
    output = Conv2D(1, (1, 1), activation="sigmoid", padding="same")(output)

    model = Model(inputs, output)
    model.compile(optimizer=optimizer, loss=loss)

    if pretrained_weights:
        model.load(pretrained_weights)

    return model



from _typeshed import Self
import pandas as pd
import datetime
from langdetect import detect
from typing import List
from collections import Counter, defaultdict
import networkx as nx
import matplotlib.pyplot as plt
from tqdm import tqdm
import spacy
import re


class TwitterClient:
    """
    Process and gain relevant information from tweets
    """

    def __init__(
        self,
        tweets: List = None,
        file_to_read: str = None,
        query_words: List = None,
        lang: str = "en",
        nlp: spacy.language.Language = None,
        ngrams: int = 2,
    ):
        """
        Initialize the class

        :param tweets: list of tweets
        :type tweets: list
        :param file_to_read: file to read
        :type file_to_read: str
        :param query_words: list of words to find cooccurrences
        :type query_words: list
        :param lang: language of tweets
        :type lang: str
        :param nlp: spacy model
        :type nlp: spacy.language.Language
        :param ngrams: ngrams to use
        :type ngrams: int

        :return: None
        """

        if file_to_read:
            self.tweets = self.read_file(file_to_read)

        else:
            self.tweets = tweets

        self.lang = lang
        self.nlp = nlp
        self.ngrams = ngrams
        self.query_words = query_words
        self.nlp = spacy.load("en_core_web_sm")

    def read_file(self, filename):
        """
        Read a file and return a list of tweets

        :param filename: file to read
        :type filename: str

        :return: list of tweets
        """

        assert (
            filename is not None and filename != ""
        ), "filename cannot be None or empty"

        tweets_from_file = pd.read_csv(filename, lineterminator="\n")
        tweets_from_file = tweets_from_file[tweets_from_file["language"] == "en"]
        tweets_from_file["datetime"] = tweets_from_file["date"].apply(
            lambda x: dateExtractor(x)
        )
        tweets_from_file["yearvalue"] = tweets_from_file["datetime"].apply(
            lambda x: x.year
        )

        return tweets_from_file

    def clean_data(self, tweets):
        """
        Clean the tweets

        :param tweets: list of tweets
        :type tweets: list

        :return: list of cleaned tweets
        """

        # define clean tweet data function
        emoji_pattern = re.compile(
            "["
            u"\U0001F600-\U0001F64F"  # emoticons
            u"\U0001F300-\U0001F5FF"  # symbols & pictographs
            u"\U0001F680-\U0001F6FF"  # transport & map symbols
            u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
            u"\U00002702-\U000027B0"
            u"\U000024C2-\U0001F251"
            u"\U00002500-\U00002BEF"  # chinese char
            u"\U0001f921-\U0001f937"
            u"\U00010000-\U0010ffff"
            u"\u2640-\u2642"
            u"\u2600-\u2B55"
            u"\u200d"
            u"\u23cf"
            u"\u23e9"
            u"\u231a"
            u"\ufe0f"  # dingbats
            u"\u3030"
            "]+",
            flags=re.UNICODE,
        )

        email_pattern = re.compile("\S+@\S+\.\S{2,3}")
        link_pattern = re.compile("https?\S+")

        tweets_cleaned = set()
        for tweet in tweets:
            try:
                lang = self.lang
                if lang == "en":
                    tweet_rep = emoji_pattern.sub(r"", tweet)  # remove emojis
                    tweet_rep = email_pattern.sub(r"", tweet_rep)  # remove emails
                    tweet_rep = link_pattern.sub(r"", tweet_rep)  # remove links
                    tweet_rep = tweet_rep.replace("’", "'")
                    tweet_rep = tweet_rep.strip()  # split tweet
                    tweets_cleaned.add(tweet_rep)
                else:
                    print("Only english tweets supported right now")
                    return []
            except:
                print("Item not processed:", tweet)
        return tweets_cleaned

    def find_cooccurrences(self, query_words):
        """
        Find cooccurrences of words in tweets

        :param query_words: list of words to find cooccurrences
        :type query_words: list

        :return: cooccurrences
        :returntype: defaultdict
        """

        self.query_words = query_words
        tweets = self.clean_data(self.tweets)

        pos_meaningful = ["NOUN", "PROPN", "ADJ", "VERB"]
        co_occurrences = defaultdict(list)  # creates a dict with list as default key

        for query in query_words:
            co_occurrences[query] = Counter()

        for doc in tqdm(self.nlp.pipe(tweets, disable=["ner"])):
            lemmas = [
                token.lemma_ for token in doc
            ]  # get list of words in doc with inflections
            for query in query_words:
                if query in lemmas:
                    # get list of meainingful word, if len of doc is greater than 1 and the word is part of the define POS list, then convert to lower case
                    meaningful_words = [
                        token.lemma_.lower()
                        for token in doc
                        if token.pos_ in pos_meaningful and len(token.text) > 1
                    ]
                    # create bigrams
                    bigrams = []
                    for i in range(len(doc) - 1):
                        if (
                            doc[i].pos_ in pos_meaningful
                            and doc[i + 1].pos_ in pos_meaningful
                        ):
                            bigram = doc[i].text + " " + doc[i + 1].text
                            bigrams.append(bigram)
                    # count number of times, meaningful phrases and bigram cooccur
                    for phrase in meaningful_words + bigrams:
                        co_occurrences[query][phrase] += 1

        self.co_occurrences = co_occurrences

        return co_occurrences

    def compute_edge_weights(self):
        """
        Compute edge weights for cooccurrences

        :return: edge weights
        :returntype: List
        """

        coocurrence_dict = self.co_occurrences

        # get total occurences
        total_occurences = 0
        for j in range(len(coocurrence_dict)):
            total_occurences = total_occurences + int(coocurrence_dict[j][1])

        # compute normalised weights
        edge_weights = []
        for i in range(len(coocurrence_dict)):
            edge_weights.append((int(coocurrence_dict[i][1]) / total_occurences) * 30)

        self.edge_weights = edge_weights

        return edge_weights

    def make_adjacency_list(self):
        """
        Make adjacency list with edge weights

        :return: adjacency list
        :returntype: List
        """

        edge_weights = (
            self.compute_edge_weights()
            if self.edge_weights is None
            else self.edge_weights
        )
        coocurrence_dict = (
            self.find_cooccurrences(self.query_words)
            if self.co_occurrences is None
            else self.co_occurrences
        )
        adjacency_list = []

        for i in range(len(coocurrence_dict)):
            edge = (coocurrence_dict[0][0], coocurrence_dict[i][0], edge_weights[i])
            adjacency_list.append(edge)

        self.adjacency_list = adjacency_list

        return adjacency_list

    def build_graph(self):
        """
        Build graph with networkx

        :return: graph
        :returntype: networkx.nx.Graph
        """

        adjacency_list = (
            self.make_adjacency_list()
            if self.adjacency_list is None
            else self.adjacency_list
        )

        network = nx.Graph()  # make empty graph

        for i in range(len(adjacency_list)):
            # build network by adding edges
            network.add_edge(
                adjacency_list[i][0], adjacency_list[i][1], weight=adjacency_list[i][2]
            )

        self.network = network

        return network

    def draw_graph(self):
        """
        Visualize the graph
        """

        network = self.network
        edge_weights = self.edge_weights

        # draw
        pos = nx.spring_layout(network)
        # colors = range(len(network))
        options = {
            "edge_color": "b",
            "node_color": "b",
            "width": 10,
            "edge_cmap": plt.cm.Blues,
            "with_labels": True,
            "font_weight": "bold",
            "width": edge_weights,
        }
        nx.draw(network, pos, **options)
        # show
        plt.show()

    @staticmethod
    def dateExtractor(x):
        """
        Extract the date from a string

        :param x: date string
        :type x: str

        :return: datetime.date2
        """

        v = ""

        try:
            v = datetime.datetime.strptime(x, "%Y-%m-%d")

        except:
            try:
                v = datetime.datetime.strptime(x, "%d-%m-%Y")
            except Exception as e:
                print(x)
                print(e)
        return v



from tqdm import tqdm
import logging
import requests
import os

# Set up logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def download_from_url(url, dst):
    """
    download dataset from a url

    :param url: url to download file
    :type url: str
    :param dst: destination place to put the file
    :type dst: str
    :returns: file size and saved datasets

    """
    req = requests.get(url, stream=True, verify=False)
    file_size = int(req.headers["Content-length"])
    if os.path.exists(dst):
        first_byte = os.path.getsize(dst)
    else:
        first_byte = 0
    pbar = tqdm(
        total=file_size,
        initial=first_byte,
        unit="B",
        unit_scale=True,
        desc=url.split("/")[-1],
    )

    with (open(dst, "wb")) as f:
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
                pbar.update(1024)

    pbar.close()
    return file_size



import tensorflow as tf
from tensorflow.keras.layers import Conv2D, UpSampling2D
from tensorflow.keras.layers import Input, concatenate
from tensorflow.keras.models import Model
from tensorflow.keras.metrics import MeanIoU
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import AveragePooling2D, LeakyReLU
from tensorflow.keras.layers import GlobalAveragePooling2D
from .loss_functions_semantic import focal_tversky_loss


def semantic_segmentation(pretrained_weights, optimizer, num_classes, input_size):
    """
    Returns a Semantic segmentation model

    :param pretrained_weights: pretrained weight matrix
    :type pretrained_weights: numpy ndarray
    :param input_size: image dimension tuple => (img_height, img_width, input_channel)
    :type input_size: tuple
    :param optimizer: optimization strategy
    :type optimizer: optimizer object
    :param num_classes: number of target classes
    :type num_classes: int
    :returns compiled Keras model
    :rtype: Keras model object
    """
    inputs = Input(input_size)

    lrelu = LeakyReLU(0.2)

    # Encoder
    conv1 = Conv2D(
        16, 1, dilation_rate=(1, 1), padding="same", kernel_initializer="he_normal"
    )(inputs)
    conv1 = BatchNormalization()(conv1)
    conv1 = lrelu(conv1)
    conv1 = Conv2D(
        16, 3, dilation_rate=(2, 2), padding="same", kernel_initializer="he_normal"
    )(conv1)
    conv1 = BatchNormalization()(conv1)
    conv1 = lrelu(conv1)

    conv2 = Conv2D(
        32, 1, dilation_rate=(1, 1), padding="same", kernel_initializer="he_normal"
    )(conv1)
    conv2 = BatchNormalization()(conv2)
    conv2 = lrelu(conv2)
    conv2 = Conv2D(
        32, 3, dilation_rate=(2, 2), padding="same", kernel_initializer="he_normal"
    )(conv2)
    conv2 = BatchNormalization()(conv2)
    conv2 = lrelu(conv2)

    conv3 = Conv2D(
        64, 1, dilation_rate=(1, 1), padding="same", kernel_initializer="he_normal"
    )(conv2)
    conv3 = BatchNormalization()(conv3)
    conv3 = lrelu(conv3)
    conv3 = Conv2D(
        64, 3, dilation_rate=(2, 2), padding="same", kernel_initializer="he_normal"
    )(conv3)
    conv3 = BatchNormalization()(conv3)
    conv3 = lrelu(conv3)

    conv4 = Conv2D(
        128, 1, dilation_rate=(1, 1), padding="same", kernel_initializer="he_normal"
    )(conv3)
    conv4 = BatchNormalization()(conv4)
    conv4 = lrelu(conv4)
    conv4 = Conv2D(
        128, 3, dilation_rate=(2, 2), padding="same", kernel_initializer="he_normal"
    )(conv4)
    conv4 = BatchNormalization()(conv4)
    conv4 = lrelu(conv4)

    # Decoder - Pyramid Module Pooling
    red = GlobalAveragePooling2D(name="red_pool")(conv4)
    red = tf.keras.layers.Reshape((1, 1, 128))(red)
    red = Conv2D(64, 1, name="red_1_by_1")(red)
    red = UpSampling2D(
        size=(400, 1920), interpolation="bilinear", name="red_upsampling"
    )(red)

    yellow = AveragePooling2D(pool_size=(2, 2), name="yellow_pool")(conv4)
    yellow = Conv2D(64, 1, name="yellow_1_by_1")(yellow)
    yellow = UpSampling2D(size=2, interpolation="bilinear", name="yellow_upsampling")(
        yellow
    )

    blue = AveragePooling2D(pool_size=(4, 4), name="blue_pool", padding="valid")(conv4)
    blue = Conv2D(64, 1, name="blue_1_by_1")(blue)
    blue = UpSampling2D(size=4, interpolation="bilinear", name="blue_upsampling")(blue)

    green = AveragePooling2D(pool_size=(4, 8), name="green_pool", padding="valid")(
        conv4
    )
    green = Conv2D(64, 1, name="green_1_by_1")(green)
    green = UpSampling2D(
        size=(4, 8), interpolation="bilinear", name="green_upsampling"
    )(green)

    merge9 = concatenate([conv4, red, yellow, blue, green])

    output = Conv2D(256, 3, padding="same", kernel_initializer="he_normal")(merge9)
    output = lrelu(output)
    output = Conv2D(3, 1, activation="softmax", padding="same")(output)

    model = Model(inputs, output)
    model.compile(
        optimizer=optimizer,
        loss=focal_tversky_loss,
        metrics=[MeanIoU(num_classes=num_classes)],
    )

    return model



import torch
import torch.nn as nn
import torch.nn.functional as F


class LabelSmoothingCrossEntropy(nn.Module):
    """
    NLL loss with label smoothing.
    Credits: timm library
    """

    def __init__(self, smoothing=0.1):
        """
        Constructor for the LabelSmoothing module.
        :param smoothing: label smoothing factor
        """
        super(LabelSmoothingCrossEntropy, self).__init__()
        assert smoothing < 1.0
        self.smoothing = smoothing
        self.confidence = 1.0 - smoothing

    def forward(self, x, target):
        logprobs = F.log_softmax(x, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss.mean()



#!/usr/bin/python3

""" geom_to_json.py
	Converts a geometry, whether a shapefile or something else, into a JSON
	construct.  This format is required by Google Earth Engine (GEE) for further
	operation such as extracting a raster from satellite images.
	https://developers.google.com/earth-engine/guides/table_upload

	This is a commen method of extracting datasets because it covers a certain
	range of time rather than a single event. The geometry bounds this collection
	to create a final raster that only contains attributes for the area of intereset
	(AOI). Because it uses GEE, the type has to be converted to JSON.
	https://developers.google.com/earth-engine/tutorials/tutorial_api_04

	NOTE ABOUT Google Earth Engine:
	This code assumes the user has the authorization and authentication to use GEE.
	https://developers.google.com/earth-engine/guides/python_install

	The code below is an extension from this Q&A link (original code):
	https://gis.stackexchange.com/questions/333791/accessing-a-shapefile-with-googleearthengine-api-invalid-geojson-geometry/334400#334400

	And syntax explanation from geopandas:
	https://geopandas.org/gallery/polygon_plotting_with_folium.html?highlight=to_json

	Libraries:
		Pandas
		Geopandas
		Google Earth Engine

	API Calls:
		gpd.GeoSeries()
		[gpd].to_json()
		ee.FeatureCollection()
		ee.Feature()

        External Dependency:
		None

"""

import pandas as pd
import geopandas as gpd
import ee


def geom_to_json(df: pd.DataFrame, i: int) -> ee.geometry.Geometry:
    """Converts shapefile or any geometry input into JSON format for GEE.

    Args:
            df:	panda dataframe containing shapefile contents
            i:	integer value indicating which row in the dataframe to process

    Returns:
            region:	the JSON string
            admin:	label for the section of the geometry, useful for file associations or names
    """

    # extract the geometry for the given AOI
    g = df.iloc[i, :]

    geom = gpd.GeoSeries(
        g["geometry"]
    )  # reflecting the new geopandas API for json conversion
    # 'geometry' is the built in label commonly used in shapefiles, see dataframe
    jsonDict = eval(geom.to_json())
    geojsonDict = jsonDict["features"][
        0
    ]  # [0]th location of the 'features' index in the JSON dictionary order

    # call to the GEE API
    region = ee.FeatureCollection(ee.Feature(geojsonDict)).geometry()

    # label the geometry for easy reference
    d = str(g["District Name"]).lower().strip()
    admin = d.replace(" ", "_")  # remove non alphanumeric character

    return region, admin


# end geom_to_json



# Adapted from: https://github.com/PRBonn/bonnetal/blob/master/train/tasks/segmentation/modules/traceSaver.py
import os
import errno
import onnx
import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class TraceSaver:
    """
    class to convert the model to ONMX

    :param model_path: Path where pytorch model is saved
    :type model_path: str
    :param new_path: Path to folder where onnx model will be saved as `model.onnx`
    :type new_path:str
    :param force_img_prop:  tuple containing (height, width, channels)
    :type force_img_prop: tuple (1,3)
    """

    def __init__(
        self, model_path: str, new_path: str, force_img_prop: tuple = (None, None, None)
    ):
        """
        :param model_path: path to model config
        :type model_path: str
        :param new_path: new_path
        :type new_path: str
        :param force_img_prop: image transform
        :type force_img_prop: Tuple
        """
        self.path = model_path
        self.new_path = new_path
        if (
            force_img_prop[0] is not None
            and force_img_prop[1] is not None
            and force_img_prop[2] is not None
        ):
            self.height = force_img_prop[0]
            self.width = force_img_prop[1]
            self.channels = force_img_prop[2]
            print("WARNING: FORCING IMAGE PROPERTIES TO")
            print(
                "(Height, Width, Channels: ({}, {}, {}))".format(
                    self.height, self.width, self.channels
                )
            )
        if os.path.exists(self.path):
            self.model = torch.load(self.path)
        else:
            print(f"No trained model found at {self.path}")
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.path)

        # CUDA speedup
        torch.backends.cudnn.benchmark = True
        self.model = self.model.to(device)

        # eval mode
        self.model.eval()
        print("Creating dummy input to profile")
        self.dummy_input = torch.randn(1, self.channels, self.height, self.width).to(
            device
        )

    def export_onnx(self, debug: bool = True):
        """
        Convert to onnx model with given height and width

        :param debug: print the onnx model graph
        :type debug: bool
        :returns:Save onnx model to folder specified in `new_path` as `model.onnx`

        """
        # create profile
        onnx_path = os.path.join(self.new_path, "model.onnx")
        with torch.no_grad():
            print("Profiling model")
            print("Saving model at ", onnx_path)
            torch.onnx.export(self.model, self.dummy_input, onnx_path)
        # check that it worked
        print("Checking that it all worked out")
        model_onnx = onnx.load(onnx_path)
        onnx.checker.check_model(model_onnx)
        # Print a human readable representation of the graph
        if debug:
            print(onnx.helper.printable_graph(model_onnx.graph))

    def export_dynamic_onnx(self, debug: bool = True):
        """
        Dynamic model onnx used to save a onnx model with different (height, weight) than
        (height, width) used for training model

        :param debug: print the onnx model graph
        :type debug: bool
        :returns:save dynamic onnx model to folder specified in `new_path` as `model_dynamic.onnx`

        """
        # convert to a dynamic ONNX traced model
        # NOTE: Dynamic width/height may not achieve the expected performance improvement
        # with some backend such as TensorRT though.
        input_names = ["input"]
        output_names = ["output"]
        dynamic_axes = {
            "input": {0: "batch_size", 2: "width", 3: "height"},
            "output": {0: "batch_size", 2: "width", 3: "height"},
        }
        # create profile
        onnx_path = os.path.join(self.new_path, "model_dynamic.onnx")
        with torch.no_grad():
            print("Profiling model")
            print("Saving model at ", onnx_path)
            torch.onnx.export(
                self.model,
                self.dummy_input,
                onnx_path,
                input_names,
                output_names,
                dynamic_axes,
            )
        # check that it worked
        print("Checking that it all worked out")
        model_onnx = onnx.load(onnx_path)
        onnx.checker.check_model(model_onnx)
        # Print a human readable representation of the graph
        if debug:
            print(onnx.helper.printable_graph(model_onnx.graph))



import cv2
import imutils
import numpy as np

import json


def hogdetection(image_path):
    """
    Histogram of gradients-based people detection function

    :param image_path: Path of the input image
    :type image_path: str
    :returns: List of (x,y,w,h) tuples for all the bbox of people detected in the image
    :rtype: list

    """
    # Initializing the HOG person detector
    hog = cv2.HOGDescriptor()
    hog.setSVMDetector(cv2.HOGDescriptor_getDefaultPeopleDetector())

    # Reading the Image
    image = cv2.imread(image_path)

    # Resizing the Image
    image = imutils.resize(image, width=min(400, image.shape[1]))

    # Detecting all the regions in the image that has a pedestrians inside it
    (regions, _) = hog.detectMultiScale(
        image, winStride=(4, 4), padding=(4, 4), scale=1.05
    )

    return regions


def MobileNet_SSDObjectDetector(image_path):
    """
    Use MobileNet_SSD Object detector to detect classes as mentioned bellow

    CLASSES = ["background", "aeroplane", "bicycle", "bird", "boat","bottle", "bus", "car",
                "cat", "chair", "cow", "diningtable","dog", "horse", "motorbike", "person",
                "pottedplant", "sheep","sofa", "train", "tvmonitor"]

    :param image_path: Path of the image
    :type image_path: str
    :returns: This is a list of tuples and tuples are in the form (label, [bbox])
    :rtype: list of tuples

    """

    with open("./components/computervision/config.json") as d:
        config = json.load(d)

    prototxt = config["MobileNet_SSDObjectDetector"]["prototxt"]
    model = config["MobileNet_SSDObjectDetector"]["model"]

    # initialize the list of class labels MobileNet SSD was trained to
    # detect, then generate a set of bounding box colors for each class
    CLASSES = [
        "background",
        "aeroplane",
        "bicycle",
        "bird",
        "boat",
        "bottle",
        "bus",
        "car",
        "cat",
        "chair",
        "cow",
        "diningtable",
        "dog",
        "horse",
        "motorbike",
        "person",
        "pottedplant",
        "sheep",
        "sofa",
        "train",
        "tvmonitor",
    ]

    # load our serialized model from disk
    print("[INFO] loading model...")
    net = cv2.dnn.readNetFromCaffe(prototxt, model)

    # by resizing to a fixed 300x300 pixels and then normalizing it
    # (note: normalization is done via the authors of the MobileNet SSD
    # implementation)

    image = cv2.imread(image_path)
    (h, w) = image.shape[:2]
    blob = cv2.dnn.blobFromImage(
        cv2.resize(image, (300, 300)), 0.007843, (300, 300), 127.5
    )

    # pass the blob through the network and obtain the detections and
    # predictions
    print("[INFO] computing object detections...")
    net.setInput(blob)
    detections = net.forward()

    result = []

    # loop over detections
    for i in np.arange(0, detections.shape[2]):
        # extract the confidence (i.e., probability) associated with the prediction
        confidence = detections[0, 0, i, 2]

        # filter out weak detections by ensuring the `confidence` is
        # greater than the minimum confidence
        if confidence > 0.2:
            # extract the index of the class label from the `detections`,
            # then compute the (x, y)-coordinates of the bounding box for
            # the object
            idx = int(detections[0, 0, i, 1])
            box = detections[0, 0, i, 3:7] * np.array([w, h, w, h])
            (startX, startY, endX, endY) = box.astype("int")
            label = "{}: {:.2f}%".format(CLASSES[idx], confidence * 100)

            result.append(tuple(label, [box]))

    return result



#!/usr/bin/python3


"""reproject_crs.py
	Reprojects a current raster file's CRS to a new user
	defined CRS.

	https://epsg.io/
	https://en.wikipedia.org/wiki/EPSG_Geodetic_Parameter_Dataset
	https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84

	This function only converts the CRS from the current value to
	a new user value, does nothing else.  The returned new dataset
	is of type xarray.core.dataarray.DataArray.

	User required to install the libraries to avoid errors.

	Libraries:
		rasterio
		rioxarray

	API Calls:
		open_rasterio()
		from_string()
		reproject()

        External Dependency:
		None
"""


import rasterio
from rasterio.crs import CRS
import rioxarray as rxr


def reproject_crs(rf: rxr, new_crs: str) -> rxr:
    """
    Reproject the raster from a current CRS to a user specified CRS.

    :param rf: (r)aster (f)ile sent for reprojection
    :param new_crs: user defined new CRS projection

    :Example:

    new_raster_object = reproject_crs(current_raster, 'EPSG:4326')

    """

    # open the raster with rioxarray
    file_crs = rxr.open_rasterio(rf, masked=True).squeeze()

    # display current CRS
    print(f"Raster file current CRS: {file_crs.rio.crs}")

    # obtain new CRS
    crs_wgs84 = CRS.from_string(new_crs)

    # Reproject the dataset with the new crs
    new_wgs84 = file_crs.rio.reproject(crs_wgs84)

    # display new CRS
    print(f"New reprojected raster file CRS: {new_wgs84.rio.crs}")

    return new_wgs84
    # end of reproject_crs



import numpy as np
from skimage.measure import compare_ssim as ssim
from math import log10, sqrt

# structural similarity index measure (SSIM)


def SSIM(predicted_image, target_image):
    """
    calculate Structural Similarity Index (SSIM) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: SSIM SCORE  which calculated between the two input images.

    """

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ssim(predicted_image, target_image, multichannel=True)
        if predicted_image.ndim == 1:
            score = ssim(predicted_image, target_image, multichannel=False)
    return score


# Peak signal-to-noise ratio (PSNR)
def PSNR(predicted_image, target_image):
    """
    calculate Peak Signal-to-Noise Ratio (PSNR) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: PSNR SCORE  which calculated between the two input images.

    """
    psnr = 0

    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate PSNR
        mse = np.mean((target_image - predicted_image) ** 2)
        if mse == 0:
            return 100
        max_pixel = 255.0
        psnr = 20 * log10(max_pixel / sqrt(mse))
    return psnr


# Mean Squared Error (MSE)
def MSE(predicted_image, target_image):
    """
    calculate Mean Squared Error (MSE) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: MSE SCORE  which calculated between the two input images.

    """
    error = np.sum(
        (target_image.astype("float") - predicted_image.astype("float")) ** 2
    )
    error /= float(target_image.shape[0] * target_image.shape[1])

    # return the MSE
    return error


# Root Mean Squared Error (RMSE)
def RMSE(predicted_image, target_image):
    """
    calculate Root Mean Squared Error (RMSE) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: RMSE SCORE  which calculated between the two input images.

    """
    error = np.sum(
        (target_image.astype("float") - predicted_image.astype("float")) ** 2
    )
    error /= float(target_image.shape[0] * target_image.shape[1])

    root_error = sqrt(error)
    # return the RMSE
    return root_error


# Intersection over Union (IoU)
def IoU(predicted_box, target_box):
    """
    calculate Intersection over Union (IoU) between Predicted Bounding-box & Target Bounding-box of objects in an image.

    :param predicted_box: Array containing predicted bounding box points [x1,x2,y1,y2]
    :type predicted_box: array
    :param target_box: Array containing target bounding box points [x1,x2,y1,y2]
    :type target_image: array
    :returns: IoU, Union and Intersection SCORES.

    """
    inter_box_top_left = [
        max(target_box[0], predicted_box[0]),
        max(target_box[1], predicted_box[1]),
    ]
    inter_box_bottom_right = [
        min(target_box[0] + target_box[2], predicted_box[0] + predicted_box[2]),
        min(target_box[1] + target_box[3], predicted_box[1] + predicted_box[3]),
    ]

    inter_box_width = inter_box_bottom_right[0] - inter_box_top_left[0]
    inter_box_height = inter_box_bottom_right[1] - inter_box_top_left[1]

    intersection = inter_box_width * inter_box_height
    union = (
        target_box[2] * target_box[3]
        + predicted_box[2] * predicted_box[3]
        - intersection
    )

    iou = intersection / union

    return iou, union, intersection


# Structural Dissimilarity (DSSIM)
def DSSIM(predicted_image, target_image):
    """
    calculate Structural Dissimilarity(DSSIM) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: DSSIM SCORE  which calculated between the two input images.

    """
    dssim_score = 0
    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ssim(predicted_image, target_image, multichannel=True)
        if predicted_image.ndim == 1:
            score = ssim(predicted_image, target_image, multichannel=False)

        dssim_score = (1 - score) / 2
    return dssim_score


# Signal to Reconstruction Error ratio (SRE)
def SRE(predicted_image, target_image):
    """

    calculate Signal to Reconstruction Error ratio (SRE) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: SRE SCORE  which calculated between the two input images.

    .. note:: You'll need to install image_similarity_measures library
    """

    import image_similarity_measures as ism

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ism.sre(target_image, predicted_image)
        if predicted_image.ndim == 1:
            score = ism.sre(target_image, predicted_image)

    return score


# Feature-based similarity index (FSIM)
def FSIM(predicted_image, target_image):
    """
    calculate Feature-based similarity index (FSIM) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: FSIM SCORE  which calculated between the two input images.

    .. note:: You'll need to install image_similarity_measures library
    """

    import image_similarity_measures as ism

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ism.fsim(target_image, predicted_image)
        if predicted_image.ndim == 1:
            score = ism.fsim(target_image, predicted_image)

    return score


# Information theoretic-based Statistic Similarity Measure (ISSM)
def ISSIM(predicted_image, target_image):
    """
    calculate Information theoretic-based Statistic Similarity Measure (ISSM) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: ISSIM SCORE  which calculated between the two input images.

    .. note:: You'll need to install image_similarity_measures library
    """

    import image_similarity_measures as ism

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ism.issim(target_image, predicted_image)
        if predicted_image.ndim == 1:
            score = ism.issim(target_image, predicted_image)

    return score


# Spectral angle mapper (SAM)
def SAM(predicted_image, target_image):
    """
    calculate Spectral angle mapper (SAM) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: SAM SCORE  which calculated between the two input images.

    .. note:: You'll need to install image_similarity_measures library
    """

    import image_similarity_measures as ism

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ism.sam(target_image, predicted_image)
        if predicted_image.ndim == 1:
            score = ism.sam(target_image, predicted_image)

    return score


# Universal image quality index (UIQ)
def UIQ(predicted_image, target_image):
    """
    calculate Universal image quality index (UIQ) between two 3 Colored Channel/Grayscale images.

    :param predicted_image: Image Data for predicted image as numpy array
    :type predicted_image: np.array
    :param target_image: Image Data for target image as numpy array
    :type target_image: np.array
    :returns: UIQ SCORE  which calculated between the two input images.

    .. note:: You'll need to install image_similarity_measures library
    """

    import image_similarity_measures as ism

    score = 0

    # check image dimensions
    if predicted_image.ndim == target_image.ndim:
        if predicted_image.ndim == 4:
            predicted_image = np.squeeze(predicted_image, axis=0)
            target_image = np.squeeze(target_image, axis=0)

        # calculate for multichannel images
        if predicted_image.ndim > 1:
            score = ism.uiq(target_image, predicted_image)
        if predicted_image.ndim == 1:
            score = ism.uiq(target_image, predicted_image)

    return score



import json
import warnings
from itertools import combinations, groupby
from typing import Dict, Iterable, List, Tuple

import numpy as np
import spacy
from spacy.language import Language
from spacy.tokens import Doc
from spacy.training import Example
import pandas as pd

Annotations = List[Tuple[int, int, str]]
AnnotationsDict = Dict[str, Iterable[int]]


class TaggedCorpus:
    def __init__(
        self,
        text: str,
        annotations: Annotations,
        tokenizer: spacy.tokenizer.Tokenizer = None,
    ) -> None:
        """
        Initialize the TaggedCorpus object.

        :param text: Input text
        :type text: str
        :param annotations: List of annotations
        :type annotations: list[tuple[int, int, str]]
        :param tokenizer: Tokenizer to use
        :type tokenizer: spacy.tokenizer.Tokenizer
        """

        self.text = text
        self.annotations = annotations
        if tokenizer is None:
            nlp = spacy.blank("en")
            # add more rules for tokenizing
            infixes = nlp.Defaults.infixes + [r"[-/=)(><]"]
            infix_regex = spacy.util.compile_infix_regex(infixes)
            nlp.tokenizer.infix_finditer = infix_regex.finditer
            self.tokenizer = nlp.tokenizer
        else:
            self.tokenizer = tokenizer
        # additional properties TODO: Could use @property decorator here
        self.arr_char_spans, self.tokens = self._get_char_spans_and_tokens(
            text=text, tokenizer=self.tokenizer
        )
        self.aligned_annotations_dict = self._align_span_dict(
            annotations=annotations, text=text, tokenizer=self.tokenizer
        )

    @staticmethod
    def _annotations2dict(annotations: Annotations) -> AnnotationsDict:
        """
        Go from annotations list to a dictionary

        :param annotations: Entity annotations [(start, stop, entity), ...]
        :type annotations: list[tuple[int, int, str]]

        :return: Dictionary of annotations {entity: [(start, stop), ...]}
        """

        s = sorted(annotations, key=lambda a: a[-1])
        grp = groupby(s, key=lambda a: a[-1])
        return {key: [val[:-1] for val in vals] for key, vals in grp}

    @staticmethod
    def _dict2annotations(annotations_dict: AnnotationsDict) -> Annotations:
        """
        Go from annotations dictionary to a list

        :param annotations_dict: Dictionary of annotations {entity: [(start, stop), ...]}
        :type annotations_dict: dict[str, list[tuple[int, int]]]
        :return: Dictionary of annotations {entity: [(start, stop), ...]}
        """

        return [
            (span[0], span[1], tag)
            for tag, spans in annotations_dict.items()
            for span in spans
        ]

    def _get_char_spans_and_tokens(
        self, text: str, tokenizer: spacy.tokenizer.Tokenizer
    ) -> Tuple[np.ndarray, List[str]]:
        """
        Get the character spans and the text for each token in the input text.

        :param text: Input text
        :type text: str
        :return: 2 components - numpy 2D array - rows are tokens, columns are start, stop character span; list of tokens
        """
        doc = tokenizer(text)
        arr = doc.to_array(["IDX", "LENGTH"])
        tokens = [token.text for token in doc]
        return np.array([arr[:, 0], arr[:, 0] + arr[:, -1]]).T, tokens

    def _align_span_dict(
        self, annotations: Annotations, text: str, tokenizer: spacy.tokenizer.Tokenizer
    ) -> Dict[str, Dict[str, List[str]]]:
        """
        Get the aligned character spans - spans start at first character index and end on last characted index + 1.
        For example, in "test-token.", `token` spans from (5,10) - as index starts at zero.

        :param annotations: Annotation list
        :type annotations: list[tuple[int, int, str]]
        :param text: Input text
        :type text: str
        :return: Dictionary of dictionaries - for each entity get the list of character spans and token sequences

        :Example:
        >>> {entity: {char_spans: [(0, 6), ...], token_seq: [ENT_1, ...]})
        """

        def _consecutive(arr: np.ndarray) -> List[Tuple[int, ...]]:
            """
            Get tuples of the consecutive integers in 1D array
            """
            idx_splits = (np.diff(arr) != 1).nonzero()[0] + 1
            arr_consec = np.split(ary=arr, indices_or_sections=idx_splits)
            return [tuple(a) for a in arr_consec]

        def _stich_spans(
            arr_spans: np.ndarray, tk_spans: np.ndarray
        ) -> List[Tuple[int, int]]:
            """
            Stich back together the spans of consecutive tokens.
            """
            first_and_last_idx = [
                (grp[0], grp[-1]) for grp in _consecutive(arr_spans.nonzero()[0])
            ]
            return [
                (token_spans[start, 0], token_spans[end, -1])
                for start, end in first_and_last_idx
            ]

        d_arr = {}
        annotations_dict = self._annotations2dict(annotations=annotations)
        token_spans, _ = self._get_char_spans_and_tokens(text=text, tokenizer=tokenizer)
        for tag, spans in annotations_dict.items():
            # span of entity must include start and end characters of token to be tagged
            is_span = tuple(
                ((token_spans[:, 0] >= span[0]) & (token_spans[:, 1] <= span[1]))
                for span in spans
            )
            # gather all the tokens that meet criteria
            is_tag = np.any(is_span, axis=0)
            arr_tagged_tokens = np.where(is_tag, tag, "O")
            # for subsequent tokens with the same entity, we want to join the spans
            aligned_char_span = (
                _stich_spans(arr_spans=is_tag, tk_spans=token_spans)
                if any(is_tag)
                else []
            )
            d_arr[tag] = {
                "char_spans": aligned_char_span,
                "token_seq": arr_tagged_tokens.tolist(),
            }
        return d_arr

    def to_tokens(self) -> List[str]:
        """
        Get the token list.

        :return: List of tokens
        """

        return self.tokens

    def to_token_char_spans(self) -> List[Tuple[int, int]]:
        """
        Get the character spans for each token.

        :return: List of character spans
        """

        return [(e[0], e[1]) for e in self.arr_char_spans.tolist()]

    def to_dict(self) -> Dict[str, List[str]]:
        """
        Get the entities in a dictionary.

        :return: Dictionary with token sequences (ex: {entity_1: [ENT1, ...], entity_2: [ENT2, ...], ...})
        """

        d = {k: v["token_seq"] for k, v in self.aligned_annotations_dict.items()}
        return d

    def to_doc(self, entity: str = None) -> spacy.tokens.Doc:
        """
        Get the entities in an annotated spaCy document.

        :param entity: Optional parameter - if set, return document annotated with only selected entity
        :type entity: str
        :return: SpaCy document annotated with entities
        """

        annotations_dict = {
            k: [(e[0], e[1]) for e in v["char_spans"]]
            for k, v in self.aligned_annotations_dict.items()
        }
        # get the character span entity annotations
        if entity is None:
            aligned_annotations = self._dict2annotations(
                annotations_dict=annotations_dict
            )
            annotations = rm_colliding(aligned_annotations)
            if sorted(annotations) != sorted(aligned_annotations):
                warnings.warn("Found overlapping spans. Keep only longest span.")
        else:
            annotations = [
                (span[0], span[1], entity) for span in annotations_dict[entity]
            ]

        # generate a blank spacy document and spans for the entity annotations
        doc = self.tokenizer(self.text)
        entity_spans = [
            doc.char_span(start, end, label=entity)
            for start, end, entity in annotations
        ]
        assert (
            None not in entity_spans
        ), f"Token with annotation {annotations[entity_spans.index(None)]} don't match token spans."
        # annotate document with entity spans
        doc.set_ents(entity_spans)
        return doc

    def to_iob_list(self, entity: str = None) -> List[str]:
        """
        Get the entities in a list of entities (IOB format).
        :param entity: Optional parameter - if set, return document annotated with only selected entity
        :type entity: str
        :return: List of strings where each string represents the entity associated with the token
        """

        doc = self.to_doc(entity=entity)
        return doc2ents(doc)

    def to_multi_iob_list(self) -> Tuple[List[Tuple[str, ...]], List[str]]:
        """
        Get all the entities in a list of entities (IOB format).

        :return: List of tuples where each tuple contains one element for each entity present in the document
        """

        all_entities = [
            self.to_iob_list(entity=entity)
            for entity in self.aligned_annotations_dict.keys()
        ]
        return list(zip(*all_entities)), list(self.aligned_annotations_dict.keys())


def get_entities(str: str) -> Annotations:
    """
    Process annotations to get in the format of (start, stop, label)

    :param str: String of annotations
    :type str: str

    :return: List of tuples of the form (start, stop, label)
    """

    d = json.loads(str)

    # check if dictionary is not empty
    if not d:
        return None

    assert isinstance(d["objects"], list), "Couldn't find `objects` list."
    return [
        (el["data"]["location"]["start"], el["data"]["location"]["end"], el["value"])
        for el in d["objects"]
    ]


def rm_groups(a: Annotations) -> Annotations:
    """
    Remove the group number info - ie: `g2_response_rate` becomes `response_rate`

    :param a: Annotations
    :type a: Annotations

    :return: Annotations with group number removed
    """

    def _rm_grp_info(s: str) -> str:
        if s.startswith("g1_") or s.startswith("g2_"):
            return s[3:]
        elif s in ["group1", "group2"]:
            return "group"
        else:
            return s

    return list({(start, stop, _rm_grp_info(ent)) for start, stop, ent in a})


def rm_colliding(x: Annotations) -> Annotations:
    """
    Remove any overlapping annotations (one token cannot have two entities)

    :param x: List of annotations
    :type x: List[Tuple[int, int, str]]

    :return: List of annotations with no overlapping annotations
    """

    def _is_overlap(a: Tuple[int, int], b: Tuple[int, int]) -> bool:
        return a[0] < b[1] and b[0] < a[1]

    def _span_to_drop(a: Tuple[int, int], b: Tuple[int, int]) -> Tuple[int, int]:
        return a if (a[1] - a[0]) < (b[1] - b[0]) else b

    def _rm_overlap(ranges: Iterable[Tuple[int, int]]):
        not_valid = [
            _span_to_drop(a, b) for a, b in combinations(ranges, 2) if _is_overlap(a, b)
        ]
        valid = [el for el in ranges if el not in not_valid]

        # check for other overlaps (in case of 3-way collisions)
        if not_valid:
            return _rm_overlap(valid)
        else:
            return valid

    # if duplicated spans then take first
    d = {(start, stop): ent for start, stop, ent in x[::-1]}
    valid_spans = _rm_overlap(d.keys())
    return [
        (start, stop, ent)
        for (start, stop), ent in d.items()
        if (start, stop) in valid_spans
    ]


def doc2ents(doc: Doc) -> List[str]:
    """
    Take a spaCy `Doc` and return a list of IOB token-based entities

    :param doc: Spacy document
    :type doc: spacy.tokens.Doc

    :return: List of entity strings
    """

    return [
        f"{token.ent_iob_}-{token.ent_type_}"
        if token.ent_type_
        else token.ent_iob_
        if token.ent_iob_
        else "O"
        for token in doc
    ]


def generate_examples(
    texts: Iterable[str], entity_offsets: Iterable[Annotations], nlp: Language
) -> List[Example]:
    """
    Generate spacy `Example` objects for training

    :param texts: List of texts to generate examples from
    :type texts: Iterable[str]
    :param entity_offsets: List of annotations for each text
    :type entity_offsets: Iterable[Annotations]
    :param nlp: Spacy model
    :type nlp: spacy.language.Language

    :return: List of spacy `Example` objects
    """

    if len(texts) != len(entity_offsets):
        raise ValueError("Lengths don't match.")
    # get docs (with predictions)
    docs = [nlp.make_doc(text) for text in texts]
    examples = [
        Example.from_dict(doc, {"entities": entity_offsets})
        for doc, entity_offsets in zip(docs, entity_offsets)
    ]
    return examples


def create_blank_nlp(train_data: pd.Series) -> spacy.Language:
    """
    Create a blank NLP model for training

    :param train_data: Dataframe of training data
    :type train_data: pd.Series

    :return: Spacy model
    """

    nlp = spacy.blank("en")

    # add more rules for tokenizing
    infixes = nlp.Defaults.infixes + [r"[-/=)(><]"]
    infix_regex = spacy.util.compile_infix_regex(infixes)
    nlp.tokenizer.infix_finditer = infix_regex.finditer

    # add ner component
    nlp.add_pipe("ner", last=True)
    ner = nlp.get_pipe("ner")

    for tags in train_data:
        for tag in tags:
            ner.add_label(tag[-1])

    return nlp



import time
import cv2
import json
import os
import numpy as np
import torch

from ..utils import Params

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class InferPytorch:
    """Pytorch inference

    Load the model with path corresponding to `save_model_path` in `param.json` file.
    :param model_dir:  Path to folder containing `params.json` file
    :type model_dir: str

    """

    def __init__(self, model_dir: str):
        """
        :param model_dir: directory from where to load model config
        :paramtype model_dir: str
        """
        json_path = os.path.join(model_dir, "params.json")
        assert os.path.isfile(
            json_path
        ), "No json configuration file found at {}".format(json_path)
        params = Params(json_path)

        # some useful data
        self.data_h, self.data_w, self.data_d = (
            params.height,
            params.width,
            params.input_channels,
        )
        self.num_classes = params.num_classes
        means, stds = params.means, params.stds
        self.means = np.array(means, dtype=np.float32)
        self.stds = np.array(stds, dtype=np.float32)

        # load the model
        try:
            self.pytorch_path = params.save_model_path
            self.model = torch.load(self.pytorch_path)
            print("Successfully Pytorch model from ", self.pytorch_path)
        except Exception as e:
            print("Couldn't load Pytorch network. Error: ", e)
            quit()

        self.model.to(device)
        # load class label mapping
        with open(model_dir + "/class_mapping.json") as fp:
            self.class_mapping = json.load(fp)

    def infer(self, bgr_img, verbose: bool = True) -> str:
        """
        Perform inference on single image

        :param bgr_img: cv2 image read using `cv2.imread`
        :type bgr_img: numpy array
        :param verbose: if True print all details such as time taken for inference, probabilities of all classes and predicted class
        :type verbose: bool
        :returns: Predicted class

        """
        # get sizes
        original_h, original_w, original_d = bgr_img.shape

        # resize
        bgr_img = cv2.resize(
            bgr_img, (self.data_w, self.data_h), interpolation=cv2.INTER_LINEAR
        )

        # check if network is RGB or mono
        if self.data_d == 3:
            # get make rgb
            if verbose:
                print("Converting bgr to rgb")
            rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB)
        elif self.data_d == 1:
            # get grayscale
            if verbose:
                print("Converting to grayscale")
            rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2GRAY)
        else:
            raise NotImplementedError(
                "Network has to have 1 or 3 channels. Anything else must be implemented."
            )
        # to tensor
        rgb_tensor = torch.from_numpy(rgb_img)
        # permute and normalize
        rgb_tensor = (rgb_tensor.float() / 255.0 - self.means) / self.stds
        rgb_tensor = rgb_tensor.permute(2, 0, 1)
        # add batch dimension
        rgb_tensor = rgb_tensor.unsqueeze(0)
        # infer
        start = time.time()
        logits = self.model(rgb_tensor.to(device))
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        argmax = logits[0].argmax(dim=0).cpu().numpy().astype(np.uint8)
        probs = torch.nn.functional.softmax(logits[0], dim=0).detach().cpu().numpy()
        time_to_infer = time.time() - start
        # time
        if verbose:
            print("Time to infer: {:.3f}s".format(time_to_infer))
            print("{:15s} | {:15s}".format("Predictions", "Probabilities"))
            print("-" * 30)
            for i in range(len(probs)):
                print(
                    "{:15s} | {:.2f}%".format(
                        self.class_mapping[str(i)], probs[i] * 100
                    )
                )
            print(
                "Top-1 Prediction: {} | Probability: {:.2f}%".format(
                    self.class_mapping[str(argmax)], probs[argmax] * 100
                )
            )
        return self.class_mapping[str(argmax)]



# Adapted from: https://github.com/PRBonn/bonnetal/blob/master/train/tasks/segmentation/modules/userTensorRT.py
import time
import cv2
import json
import os
import numpy as np

import tensorrt as trt
import pycuda.autoinit
import pycuda.driver as cuda

from .trtCalibINT8 import EntropyCalibrator
from ..utils import Params

TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE)


class InferTensorRT:
    """Convert to TensorRT model and use it for inference

    :param model_dir: Path to parent folder containing onnx model stored inside `models` folder with name `model.onnx`
    :type model_dir: str
    :param mode: Mode for tensorrt conversion [fp16, int8, both]
    :type mode: str
    :param calib_dir: Path to folder containing sample images that will be used for int8 calibration.[Default = None]
    :type calib_dir:str

    """

    def __init__(self, model_dir: str, mode: str, calib_dir: str = None):
        """
        :param model_dir: directory from where to load model params
        :type model_dir: str
        :param mode: model mode
        :type mode: str
        :param calib_dir: calibration directory
        :type calib_dir: str
        """
        self.model_path = os.path.join(model_dir, "models/")
        json_path = os.path.join(model_dir, "params.json")
        assert os.path.isfile(
            json_path
        ), "No json configuration file found at {}".format(json_path)
        params = Params(json_path)

        data_h, data_w, data_c = params.height, params.width, params.input_channels
        self.num_classes = params.num_classes
        means, stds = params.means, params.stds
        self.means = np.array(means, dtype=np.float32)
        self.stds = np.array(stds, dtype=np.float32)

        # load class label mapping
        with open(model_dir + "/class_mapping.json") as fp:
            self.class_mapping = json.load(fp)

        network_flags = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        self.engine_serialized_path = os.path.join(
            self.model_path, f"model_mode_{mode}.trt"
        )

        calib_images = os.listdir(calib_dir)
        calib_dataset = [os.path.join(calib_dir, x) for x in calib_images]

        # Determine dimensions and create CUDA memory buffers
        # to hold host inputs/outputs.
        self.data_h, self.data_w, self.data_d = data_h, data_w, data_c
        d_input_size = self.data_h * self.data_w * self.data_d * 4
        d_output_size = self.num_classes * 4
        # Allocate device memory for inputs and outputs.
        self.d_input = cuda.mem_alloc(d_input_size)
        self.d_output = cuda.mem_alloc(d_output_size)
        # Create a stream in which to copy inputs/outputs and run inference.
        self.stream = cuda.Stream()

        # make a builder
        builder = trt.Builder(TRT_LOGGER)

        # support int 8 calibration?
        print("Platform has int8 mode: ", builder.platform_has_fast_int8)
        # the calibrator
        print("Trying to calibrate int8 from list of images")
        self.engine = None  # force rebuild of engine
        calib = None
        if mode == "int8" or mode == "both":
            calib = EntropyCalibrator(
                model_path=self.model_path,
                calibration_files=calib_dataset,
                batch_size=1,
                h=self.data_h,
                w=self.data_w,
                means=self.means,
                stds=self.stds,
            )

        # architecture definition from onnx if no engine is there
        # get weights?
        if self.engine is None:
            try:
                # basic stuff for onnx parser
                model_path = os.path.join(self.model_path, "model.onnx")
                network = builder.create_network(network_flags)
                onnxparser = trt.OnnxParser(network, TRT_LOGGER)
                model = open(model_path, "rb")
                onnxparser.parse(model.read())
                print("Successfully ONNX weights from ", model_path)
            except Exception as e:
                print("Couldn't load ONNX network. Error: ", e)
                quit()

            print("Wait while tensorRT profiles the network and build engine")
            # trt parameters
            try:
                builder.max_batch_size = 1
                builder.max_workspace_size = 8000000000
                builder.int8_mode = False
                builder.fp16_mode = False
                if mode == "int8":
                    builder.int8_calibrator = calib
                    builder.int8_mode = True
                if mode == "both":
                    builder.int8_calibrator = calib
                    builder.fp16_mode = True
                    builder.int8_mode = True
                if mode == "fp16":
                    builder.fp16_mode = True
                print("-" * 80)
                print(f"Performing {mode} conversion")
                print("-" * 80)
                print("Calling build_cuda_engine")
                self.engine = builder.build_cuda_engine(network)
                assert self.engine is not None
            except Exception as e:
                print("Failed creating self.engine for TensorRT. Error: ", e)
                quit()
            print("Done generating tensorRT self.engine.")

            # serialize for later
            print(
                "Serializing tensorRT self.engine for later (for example in the C++ interface)"
            )
            try:
                self.serialized_engine = self.engine.serialize()
                with open(self.engine_serialized_path, "wb") as f:
                    f.write(self.serialized_engine)
            except Exception as e:
                print(
                    "Couldn't serialize self.engine. Not critical, so I continue. Error: ",
                    e,
                )
        else:
            print("Successfully opened self.engine from inference directory.")
            print(
                "WARNING: IF YOU WANT TO PROFILE FOR THIS COMPUTER DELETE model.trt FROM THAT DIRECTORY"
            )

        # create execution context
        self.context = self.engine.create_execution_context()

    def infer(self, bgr_img, topk: int = 1, verbose: bool = True) -> tuple:
        """
        Inference on single image

        :param bgr_img: cv2 image read using `cv2.imread`
        :type bgr_img: numpy array
        :param topk: top-k classes to extract
        :type topk: int
        :param verbose: if true, print time required for inference
        :type verbose: bool
        :returns: list of top-k class strings and corresponding class indexes

        """
        # get sizes
        original_h, original_w, original_d = bgr_img.shape
        # resize
        bgr_img = cv2.resize(
            bgr_img, (self.data_w, self.data_h), interpolation=cv2.INTER_LINEAR
        )

        # check if network is RGB or mono
        if self.data_d == 3:
            # get make rgb
            if verbose:
                print("Converting bgr to rgb")
            rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB)
        elif self.data_d == 1:
            # get grayscale
            if verbose:
                print("Converting to grayscale")
            rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2GRAY)
        else:
            raise NotImplementedError(
                "Network has to have 1 or 3 channels. Anything else must be implemented."
            )

        # permute and normalize
        rgb_tensor = (rgb_img.astype(np.float32) / 255.0 - self.means) / self.stds
        rgb_tensor = np.transpose(rgb_tensor, (2, 0, 1))

        # add batch dimension
        rgb_tensor = rgb_tensor[np.newaxis, ...]

        # placeholders
        h_input = np.ascontiguousarray(rgb_tensor, dtype=np.float32)
        h_output = np.empty(self.num_classes, dtype=np.int32, order="C")

        # infer
        start = time.time()

        # Transfer input data to the GPU.
        cuda.memcpy_htod_async(self.d_input, h_input, self.stream)
        # Run inference.
        self.context.execute_async(
            bindings=[int(self.d_input), int(self.d_output)],
            stream_handle=self.stream.handle,
        )
        # Transfer predictions back from the GPU.
        cuda.memcpy_dtoh_async(h_output, self.d_output, self.stream)
        # Synchronize the stream
        self.stream.synchronize()

        logits = h_output.reshape((self.num_classes))
        max_k = np.argpartition(logits, -topk)[-topk:]
        time_to_infer = time.time() - start

        # print result, and put in output lists
        classes = []
        classes_str = []
        if verbose:
            print("Time to infer: {:.3f}s".format(time_to_infer))
        for i, idx in enumerate(max_k):
            class_string = self.class_mapping[str(idx)]
            classes.append(idx)
            classes_str.append(class_string)
            if verbose:
                print("[{}]: {}, {}".format(i, idx, class_string))
        return classes, classes_str



import imghdr
from pathlib import Path
import gdal


class SatelliteImage:
    """
    SatelliteImage class is used to preprocess satellite images

    :param path: Path of the input .tif image
    :type path: str
    :param parts: Splits the single image into smaller images with a total of (parts * parts) images returned
    :type parts: int

    :example:

    >>> si = SatelliteImage(
    >>>        path="sample.tif", parts=5
    >>> )
    >>> si.split()
    """

    def __init__(self, path: str, parts: int) -> None:
        self.path = path
        assert self.path != "", "Path cannot be empty"
        assert imghdr.what(self.path) == "tiff", "You need to pass a .tif image"
        self.dem = gdal.Open(self.path)
        self.gt = self.dem.GetGeoTransform()
        self.parts = parts
        assert self.parts != 0, "Parts needs to be greater than 0"
        self.dir = Path(self.path).parent

    def split(self) -> None:
        """Splits a single .tif image into multiple parts in the same directory

        This makes it easier for
        visualising and storing in memory compared to having a single large
        image
        """
        x_min = self.gt[0]
        y_max = self.gt[3]
        res = self.gt[1]
        x_len = res * self.dem.RasterXSize
        y_len = res * self.dem.RasterYSize
        x_div = self.parts
        y_div = self.parts
        x_size = x_len / x_div
        y_size = y_len / y_div
        x_steps = [x_min + x_size * i for i in range(x_div + 1)]
        y_steps = [y_max - y_size * i for i in range(y_div + 1)]

        for i in range(x_div):
            for j in range(y_div):
                xmin = x_steps[i]
                xmax = x_steps[i + 1]
                ymax = y_steps[j]
                ymin = y_steps[j + 1]
                print(f"xmin: {xmin}\nxmax: {xmax}\nymin: {ymin}\nymax: {ymax}")
                gdal.Warp(
                    f"{self.dir}/dem-{i}-{j}.tif",
                    self.dem,
                    outputBounds=(xmin, ymin, xmax, ymax),
                    dstNodata=-9999,
                )
        self.dem = None



import numpy as np
from typing import Tuple, Sequence, Any
import importlib.util
from PIL import Image
import cv2
import json

try:
    import torch
except:
    print(
        "some of the functions use torch, if you use those, make sure torch is installed"
    )


class Params:
    """
    Class that loads hyperparameters from a json file.

    Example:

    >>> params = Params(json_path)
    >>> print(params.learning_rate)
    >>> params.learning_rate = 0.5  # change the value of learning_rate in params

    """

    def __init__(self, json_path: str):
        """
        :param json_path: patht to save files in
        :paramtype json_path: str:
        """
        self.update(json_path)

    def save(self, json_path: str):
        """
        Saves parameters to json file

        :param json_path: patht to save files in
        :paramtype json_path: str:
        """
        with open(json_path, "w") as f:
            json.dump(self.__dict__, f, indent=4)

    def update(self, json_path: str):
        """
        Loads parameters from json file

        :param json_path: patht to save files in
        :paramtype json_path: str:
        """
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    def __str__(self) -> str:
        """
        String presentation of the class Params
        """
        return str(self.__dict__)

    @property
    def dict(self):
        """
        Gives dict-like access to Params instance by `params.dict['learning_rate']`
        """
        return self.__dict__


# ----------------------------functions to check environment-------------------------------


def check_imshow():
    """
    Check if environment supports image displays

    :return: Return true of false if you can display image using opencv or not
    :rtype: Boolean
    """
    try:
        assert not isdocker(), "cv2.imshow() is disabled in Docker environments"
        cv2.imshow("test", np.zeros((1, 1, 3)))
        cv2.waitKey(1)
        cv2.destroyAllWindows()
        cv2.waitKey(1)
        return True
    except Exception as e:
        print(
            f"WARNING: Environment does not support cv2.imshow() or PIL Image.show() image displays\n{e}"
        )
        return False


# --------------------------functions related to image, bbox and masks--------------------------------


def zoom_to_fill(image, mask, padding):
    """
    Use the mask to make the object as the center object of the image with paddings

    :param image: image from which the object is taken out of
    :type image:  numpy.array
    :param mask: 2d mask array
    :type mask:  numpy.array
    :param padding: add black pixel padding around the image
    :type padding: int
    :return: Image array
    :rtype: numpy.array

    """
    assert isinstance(padding, int), "Only integers are allowed in padding"

    (y, x) = np.where(mask == 255)
    (topy, topx) = (np.min(y), np.min(x))
    (bottomy, bottomx) = (np.max(y), np.max(x))
    image = image[
        topy - padding : bottomy + padding, topx - padding : bottomx + padding
    ]

    return image


def translate_boxes(
    boxes: Sequence[Sequence[Any]], left: Tuple[int], top: Tuple[int]
) -> Sequence[Sequence[Any]]:
    """
    Translates bouding box by moving its cooridantes left-wise by `left` pixels and top-wise by `top` pixels.

    :param boxes: list of box coordinates <label> <left> <top> <right> <bottom>
    :type boxes: sequence
    :param left: Number of pixels to subtract from horizontal coordinates of the bouding box.Moving bouding box to the left is done with *left* > 0, and moving to the right with *left* < 0
    :type left: int
    :param top: Number of pixels to subtract from vertical coordinates of the bouding box. Moving bouding box top is done with *top* > 0, and moving it down with *top* < 0.
    :type top: int
    :returns: list of new box coordinates
    :rtype: list,same as input

    """

    transl_boxes = []
    for box in boxes:
        label, xmin, ymin, xmax, ymax = box
        transl_box = [
            label,
            int(xmin - left[0]),
            int(ymin - top[1]),
            int(xmax - left[0]),
            int(ymax - top[1]),
        ]
        transl_boxes.append(transl_box)

    return transl_boxes


def load_image(path: str):
    """
    Load an image at `path` using PIL and return the Image object and the width and height

    :param path: path where the image to be loaded is
    :type path: str

    :returns:
        - (PIL.Image.Image): Image object of the image
        - (int): width of the image
        - (int): height of the image

    """

    try:
        img = Image.open(path)
    except:
        print(f"Image at path {path} cannot be loaded")

    return img, img.size[0], img.size[1]


# ----------------------------functions related to type converstions of data----------------------------------


def xyxy2xywh(x):
    """
    Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right

    :param x: array of bbox

    :returns:
        - y: array of bbox
    """

    y = x.clone()
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height

    return y


def xywh2xyxy(x):
    """
    Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right

    :param x: array of bbox

    :returns:
        - y: array of bbox
    """

    y = x.clone()
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y

    return y


# ------------------functions to calculate matrices----------------------------------------------


def box_iou(box1, box2):
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.

    :param box1: (Tensor[N, 4])
    :param box2: (Tensor[M, 4])
    :returns:
        - iou (Tensor[N, M]): the NxM matrix containing the pairwise IoU values for every element in boxes1 and boxes2
    """
    assert importlib.util.find_spec("torch"), "torch is not installed"

    def box_area(box):
        # box = 4xn
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1.T)
    area2 = box_area(box2.T)

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    inter = (
        (
            torch.min(box1[:, None, 2:], box2[:, 2:])
            - torch.max(box1[:, None, :2], box2[:, :2])
        )
        .clamp(0)
        .prod(2)
    )
    return inter / (
        area1[:, None] + area2 - inter
    )  # iou = inter / (area1 + area2 - inter)


def ap_per_class(tp, conf, pred_cls, target_cls):
    """
    Compute the average precision, given the recall and precision curves.
    Source: https://github.com/rafaelpadilla/Object-Detection-Metrics.

    :param tp:  True positives (nparray, nx1 or nx10).
    :param conf:  Objectness value from 0-1 (nparray).
    :param pred_cls:  Predicted object classes (nparray).
    :param target_cls:  True object classes (nparray).
    :returns:
        - The average precision as computed in py-faster-rcnn.
    """

    # Sort by objectness
    i = np.argsort(-conf)
    tp, conf, pred_cls = tp[i], conf[i], pred_cls[i]

    # Find unique classes
    unique_classes = np.unique(target_cls)
    nc = unique_classes.shape[0]  # number of classes, number of detections

    # Create Precision-Recall curve and compute AP for each class
    px, py = np.linspace(0, 1, 1000), []  # for plotting
    ap, p, r = np.zeros((nc, tp.shape[1])), np.zeros((nc, 1000)), np.zeros((nc, 1000))
    for ci, c in enumerate(unique_classes):
        i = pred_cls == c
        n_l = (target_cls == c).sum()  # number of labels
        n_p = i.sum()  # number of predictions

        if n_p == 0 or n_l == 0:
            continue
        else:
            # Accumulate FPs and TPs
            fpc = (1 - tp[i]).cumsum(0)
            tpc = tp[i].cumsum(0)

            # Recall
            recall = tpc / (n_l + 1e-16)  # recall curve
            r[ci] = np.interp(
                -px, -conf[i], recall[:, 0], left=0
            )  # negative x, xp because xp decreases

            # Precision
            precision = tpc / (tpc + fpc)  # precision curve
            p[ci] = np.interp(-px, -conf[i], precision[:, 0], left=1)  # p at pr_score

            # AP from recall-precision curve
            for j in range(tp.shape[1]):
                ap[ci, j], mpre, mrec = compute_ap(recall[:, j], precision[:, j])

    # Compute F1 (harmonic mean of precision and recall)
    f1 = 2 * p * r / (p + r + 1e-16)
    i = f1.mean(0).argmax()  # max F1 index
    return p[:, i], r[:, i], ap, f1[:, i], unique_classes.astype("int32")


def compute_ap(recall, precision):
    """
    Compute the average precision, given the recall and precision curves

    :param recall: The recall curve (list)
    :param precision: The precision curve (list)
    :returns:
        - Average precision, precision curve, recall curve
    """

    # Append sentinel values to beginning and end
    mrec = np.concatenate(([0.0], recall, [recall[-1] + 0.01]))
    mpre = np.concatenate(([1.0], precision, [0.0]))

    # Compute the precision envelope
    mpre = np.flip(np.maximum.accumulate(np.flip(mpre)))

    # Integrate area under curve
    method = "interp"  # methods: 'continuous', 'interp'
    if method == "interp":
        x = np.linspace(0, 1, 101)  # 101-point interp (COCO)
        ap = np.trapz(np.interp(x, mrec, mpre), x)  # integrate
    else:  # 'continuous'
        i = np.where(mrec[1:] != mrec[:-1])[0]  # points where x axis (recall) changes
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])  # area under curve

    return ap, mpre, mrec


def set_logger(log_path: str):
    """
    Sets the logger to log info in terminal and file `log_path`.
    In general, it is useful to have a logger so that every output to the terminal is saved
    in a permanent file. Here we save it to `model_dir/train.log`.
    Example:
    ```
    logging.info("Starting training...")
    ```
    :param log_path: where to log
    """
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    if not logger.handlers:
        # Logging to a file
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s:%(levelname)s: %(message)s")
        )
        logger.addHandler(file_handler)

        # Logging to console
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter("%(message)s"))
        logger.addHandler(stream_handler)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="omdenalore",
    version="0.2.1",
    author="Omdena Collaborators",
    author_email="kaushal@omdena.com",
    description="AI for Good library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://omdena.com/omdenalore/",
    project_urls={
        "Documentation": "https://omdenaai.github.io/",
        "OmdenaLore Article": "https://omdena.com/blog/worlds-largest-ai4good-python-library-omdenalore/",
        "Apply to be a contributor": "https://airtable.com/shrue8xCVpjPvDD1g",
        "Become a partner": "https://bit.ly/omdenalore_company",
        "Join discord": "https://discord.gg/s9h7JvJzUx"

    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages()
)



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.regularizers import l2
from tensorflow.keras.metrics import MeanIoU
from tensorflow.keras import backend as K
from tensorflow.keras.layers import LeakyReLU

# Focal Tversky_loss
def class_tversky(y_true, y_pred):
    """
    Returns Tversky Class for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    smooth = 1

    y_true = K.permute_dimensions(y_true, (3, 1, 2, 0))
    y_pred = K.permute_dimensions(y_pred, (3, 1, 2, 0))

    y_true_pos = K.batch_flatten(y_true)
    y_pred_pos = K.batch_flatten(y_pred)
    true_pos = K.sum(y_true_pos * y_pred_pos, 1)
    false_neg = K.sum(y_true_pos * (1 - y_pred_pos), 1)
    false_pos = K.sum((1 - y_true_pos) * y_pred_pos, 1)
    alpha = 0.7
    return (true_pos + smooth) / (
        true_pos + alpha * false_neg + (1 - alpha) * false_pos + smooth
    )


def focal_tversky_loss(y_true, y_pred):
    """
    Returns focal Tversky loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    pt_1 = class_tversky(y_true, y_pred)
    gamma = 0.75
    return K.sum(K.pow((1 - pt_1), gamma))


# Dice Loss
smooth = 1.0


def dice_coef(y_true, y_pred):
    """
    Returns dice coefficent for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (2.0 * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)


def dice_coef_loss(y_true, y_pred):
    """
    Returns dice_coef_loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    return 1 - dice_coef(y_true, y_pred)


def generalized_dice_coefficient(y_true, y_pred):
    """
    Returns generalized_dice_coefficient for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    smooth = 1.0
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    score = (2.0 * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
    return score


def dice_loss(y_true, y_pred):
    """
    Returns dice_loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    loss = 1 - generalized_dice_coefficient(y_true, y_pred)
    return loss


def confusion(y_true, y_pred):
    """
    Returns confusion matrix for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    smooth = 1
    y_pred_pos = K.clip(y_pred, 0, 1)
    y_pred_neg = 1 - y_pred_pos
    y_pos = K.clip(y_true, 0, 1)
    y_neg = 1 - y_pos
    tp = K.sum(y_pos * y_pred_pos)
    fp = K.sum(y_neg * y_pred_pos)
    fn = K.sum(y_pos * y_pred_neg)
    prec = (tp + smooth) / (tp + fp + smooth)
    recall = (tp + smooth) / (tp + fn + smooth)
    return prec, recall


def true_positive(y_true, y_pred):
    """
    Returns True positives for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    smooth = 1
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    y_pos = K.round(K.clip(y_true, 0, 1))
    tp = (K.sum(y_pos * y_pred_pos) + smooth) / (K.sum(y_pos) + smooth)
    return tp


def true_negative(y_true, y_pred):
    """
    Returns true negetive for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    smooth = 1
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    y_pred_neg = 1 - y_pred_pos
    y_pos = K.round(K.clip(y_true, 0, 1))
    y_neg = 1 - y_pos
    tn = (K.sum(y_neg * y_pred_neg) + smooth) / (K.sum(y_neg) + smooth)
    return tn


def tversky_index(y_true, y_pred):
    """
    Returns Tversky index for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    y_true_pos = K.flatten(y_true)
    y_pred_pos = K.flatten(y_pred)
    true_pos = K.sum(y_true_pos * y_pred_pos)
    false_neg = K.sum(y_true_pos * (1 - y_pred_pos))
    false_pos = K.sum((1 - y_true_pos) * y_pred_pos)
    alpha = 0.7
    return (true_pos + smooth) / (
        true_pos + alpha * false_neg + (1 - alpha) * false_pos + smooth
    )


def tversky_loss(y_true, y_pred):
    """
    Returns Tversky loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    return 1 - tversky_index(y_true, y_pred)


def focal_tversky(y_true, y_pred):
    """
    Returns focal Tversky loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    pt_1 = tversky_index(y_true, y_pred)
    gamma = 0.75
    return K.pow((1 - pt_1), gamma)


def log_cosh_dice_loss(y_true, y_pred):
    """
    Returns log_cosh_dice loss for truth vs prediction values

    :param y_true: A tensor of the same shape as `y_pred`
    :type y_true: tensor
    :param y_pred: A tensor resulting from a softmax
    :type y_pred: tensor
    :return: Output tensor.
    """
    x = dice_loss(y_true, y_pred)
    return tf.math.log((tf.exp(x) + tf.exp(-x)) / 2.0)



"""
OmdenaLore module containing useful functions on image processing and image features
"""

import cv2
import glob
import mahotas
import numpy as np
import imutils
from matplotlib import pyplot as plt
import numpy as np


def surf_features(image_path):
    """
    detect SURF features from an image path

    :param image_path: Path of the input image
    :type image_path: str
    :returns: SURF keypoints detected from an image

    """

    img = cv2.imread(image_path)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    surf = cv2.xfeatures2d.SURF_create(400)
    kp, des = surf.detectAndCompute(gray, None)

    return kp


def sift_features(image_path, grayscale=True):
    """
    detect SIFT features from an image path

    :param image_path: Path of the input image
    :type image_path: str
    :param grayscale: If you want to convert your image to grayscale before calculating features
    :type grayscale: boolean
    :returns: SIFT keypoints detected from an image

    """
    im = cv2.imread(image_path)

    if grayscale:
        gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
        sift = cv2.xfeatures2d.SIFT_create()
        kp = sift.detect(gray, None)
    else:
        sift = cv2.xfeatures2d.SIFT_create()
        kp = sift.detect(im, None)

    return kp


def brief_features(image_path):
    """
    detect BRIEF features from an image path

    :param image_path: Path of the input image
    :type image_path: str
    :returns: BRIEF keypoints detected from an image
    """
    im = cv2.imread(image_path)
    star = cv2.xfeatures2d.StarDetector_create()
    brief = cv2.xfeatures2d.BriefDescriptorExtractor_create()
    kps = star.detect(im, None)
    kp, des = brief.compute(im, kps)
    return kp


def haralicks_features(image_path):
    """
    Detects Harlicks features from images in meaned four directions
    inside an folder with certain extentions
    and returns array of retrieved features

    :param image_path: Path of the folder which contains the png images
    :type image_path: string

    :returns: array of extracted features. The content of array is in form - (image_name , features)
    """
    assert isinstance(image_path, str), "the path should be string"

    accepted_image_extentions = ["jpeg", "png"]

    data = []
    label = []

    for imagePaths in glob.glob(image_path + "/*"):

        # check to see if a image is of acceptable format or not
        for i in range(len(accepted_image_extentions)):
            if not imagePaths.endswith(accepted_image_extentions[i]):
                print(
                    "Extention not accepted, please add this extention to the accepted extention list"
                )
                print(imagePaths)
                return None

        # load the image, convert it into greyscalse,
        image = cv2.imread(imagePaths)
        image = cv2.cvtColor(image, cv2.COLOR_BG2GRAY)

        # extract image name form the path
        label.append(imagePaths[imagePaths.rfind("/") + 1 :])

        # extract haralicks texture features in 4 directions,
        # take mean if each direction
        features = mahotas.features.haralick(image).mean(axis=0)

        # update the data with features
        data.append(features)

    return zip(label, data)


def discribe_with_zernike_moments(image_path):
    """
    Caclulates Zernike moments of images inside a folder.
    Returns list of features and cooresponding image names
    Zernike moments are great to discribing shapes of objects.

    :param image_path: Path of images
    :type image_path: string

    :returns: return a tuple of the contours and shapes (cnts, shapefeatures)
    """

    assert isinstance(image_path, str), "the path should be string"
    accepted_image_extentions = ["jpeg", "png"]
    shapeFeatures = []

    if not image_path.endswith(accepted_image_extentions[i]):
        print(
            "Extention not accepted, please add this extention to the accepted extention list"
        )
        print(image_path)
        return None

    # load the image, convert it into greyscalse,
    image = cv2.imread(image_path)
    gray = cv2.cvtColor(image, cv2.COLOR_BG2GRAY)
    blurred = cv2.GaussianBlur(gray, (13, 13), 9)
    thresh = cv2.threshold(blurred, 50, 255, cv2.THRESH_BINARY)[1]

    # perform a series of dilations and erosions to close holes
    # in the shapes
    thresh = cv2.dilate(thresh, None, iterations=4)
    thresh = cv2.erode(thresh, None, iterations=2)

    # detect contours in the edge map
    cnts = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = imutils.grab_contours(cnts)
    # loop over the contours
    for c in cnts:
        # create an empty mask for the contour and draw it
        mask = np.zeros(image.shape[:2], dtype="uint8")
        cv2.drawContours(mask, [c], -1, 255, -1)

        # extract the bounding box ROI from the mask
        (x, y, w, h) = cv2.boundingRect(c)
        roi = mask[y : y + h, x : x + w]

        # compute Zernike Moments for the ROI and update the list
        # of shape features
        features = mahotas.features.zernike_moments(
            roi, cv2.minEnclosingCircle(c)[1], degree=8
        )
        shapeFeatures.append(features)

    # return a tuple of the contours and shapes
    return (cnts, shapeFeatures)


def find_contours(image_path, show=False):
    """
    Detect contours from an image path

    :param image_path: Path of the input image
    :type image_path: str
    :param show: Whether to show the contours on a plot using matplotlib
    :type show: boolean
    :returns: contours detected from an image
    """

    assert type(image_path) == str, "Param image_path must be of type string"
    assert type(show) == bool, "Param show must be of type bool"

    im = cv2.imread(image_path)

    # convert to RGB
    rgb_image = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    # convert to grayscale
    gray = cv2.cvtColor(im, cv2.COLOR_RGB2GRAY)

    # create a binary thresholded image
    _, binary = cv2.threshold(gray, 225, 255, cv2.THRESH_BINARY_INV)

    gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)

    contours, _ = cv2.findContours(binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    if show:
        image = cv2.drawContours(rgb_image, contours, -1, (0, 255, 0), 3)
        plt.imshow(image)
        plt.show()

    return contours


def get_hough_lines(image_path, show=False):
    """
    Detect lines from an image path

    :param image_path: Path of the input image
    :type image_path: str
    :param show: Whether to show the lines on a plot using matplotlib
    :type show: boolean
    :returns: lines detected from an image
    """

    assert type(image_path) == str, "Param image_path must be of type string"
    assert type(show) == bool, "Param show must be of type bool"

    img = cv2.imread(image_path)

    grayscale = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # perform edge detection
    edges = cv2.Canny(grayscale, 30, 100)

    # detect lines in the image using hough lines technique
    lines = cv2.HoughLinesP(edges, 1, np.pi / 180, 60, np.array([]), 50, 5)

    if show:
        # iterate over the output lines and draw them
        for line in lines:
            for x1, y1, x2, y2 in line:
                cv2.line(img, (x1, y1), (x2, y2), (20, 220, 20), 3)

        # show the image
        plt.imshow(img)
        plt.show()

    return lines



import requests
from bs4 import BeautifulSoup


def getWebPage(topic_url):
    """
    Get the content of the page you want to scrape
    :param arg1: The URL of the page you want to scrape
    :type arg1: str
    :return : the status code of the response, content from this url or an assertion error
    """

    response = requests.get(topic_url)

    try:
        assert response.status_code in [200, 201, 202, 203, 204, 205, 207, 208, 226]
        return response.status_code, response.text
    except AssertionError:
        print(
            f"Problem with your url or your request. Try to check it.status_code: {response.status_code}"
        )
        return None, None


def putContentFile(filename, contents, encoding="utf-8"):
    """
    Get a new html file where the content of a web page is written
    :param arg1: name of the html file in which you want to put the content
    :param arg2: the contents we will write in the file
    :param arg3: the encoding of the file
    :type arg1: str
    :type arg2: object get from response object
    :type arg3: str
    """
    with open(filename, mode="w", encoding=encoding) as file:
        file.write(contents)


def openFile(filename):
    """
    Open the content scraped from a website
    :param arg1: the name of an existing file
    :return: the content of the file
    """
    with open(filename, "r") as f:
        html_source = f.read()
        return html_source


def get_tag(doc, tag_name):
    """
    Get a tag and it's the content.
    :param arg1: the reference for the BeautifulSoup object created
    :param arg2: the name of the tag
    :type arg1: BeautifulSoup object
    :type arg2: str
    :return : the tag and it's content
    """
    return doc.find(tag_name)


def get_groups_of_tag(doc, tag_name):
    """
    Get groups of same tag in the page.
    :param arg1: the reference for the BeautifulSoup object created
    :param arg2: the name of the tag
    :type arg1: BeautifulSoup object
    :type arg2: str
    :return : a list of same tag and their content
    """
    return doc.findAll(tag_name)


def find_tag_with_attributes(doc, tag_name, attr_dict):
    """
    Get groups of same tag in the page.
    :param arg1: the reference for the BeautifulSoup object created
    :param arg2: the name of the tag
    :param arg3: A dictionnary of attribute-values
    :type arg1: BeautifulSoup object
    :type arg2: str
    :type arg3: dictionnary
    :return : a list of same tag having the same attributes defined in the attr_dict
    """
    return doc.findAll(tag_name, attr_dict)


def findChildren(beautyobject, tagvalue, childvalue, classname, deeper=False):
    """
    Get for a tag and its child and little child.
    :param arg1: the beautifulsoup object linked to the data scraped
    :param arg2: the tag for which you search for its child
    :param arg3: the tag for child you are searching for
    :param arg4: the class for father tag
    :param arg5: the deeper you want to go in the tree
    :type arg1: BeautifulSoup object
    :type arg2: str
    :type arg3: str
    :type arg4: str
    :type arg5: boolean
    :return: a list of children of the tag
    """
    return beautyobject.find(tagvalue, class_=classname).find_all(
        childvalue, recursive=deeper
    )


def findTagAttr(doc, tagname, classname, *attributes):
    """
    Get a list of dictionnaries representing for each tag of class classname theirs attributes and their values
      :param arg1: the reference for the BeautifulSoup object created
      :param arg2: the name of the  tag
      :param arg3: the name of the class of the tag
      :param arg4: the list of attributes we need
      :type arg1: BeautifulSoup object
      :type arg2: str
      :type arg3: str
      :type arg4: list of variables arguments
      :return : a list of  dictionnaries where the attributes names are the keys and their values the corresponding values
    """
    attrList = []
    tags = find_tag_with_attributes(doc, tagname, {"class": classname})
    for tag in tags:

        listval = []
        for attr in attributes:
            listval.append(tag[attr])
            # Add the text of the tag

        dictattr = dict(zip(list(attributes), listval))
        # Add the text of the tag if it exists
        if tag.text != "":
            dictattr["text"] = tag.text.strip()
        attrList.append(dictattr)
    return attrList


def write_csv(items, path):
    """
    Write a lits of dictionnary of attributes values pairs in a CSV file
    :arg1: the list of dictionnaries
    :arg2: the path to the data
    :type arg1: list
    :type arg2: str
    :return : a csv file containing the data
    """
    # Open the file in write mode
    with open(path, "w") as f:
        # Return if there's nothing to write
        if len(items) == 0:
            return

        # Write the headers in the first line
        headers = list(items[0].keys())
        f.write(",".join(headers) + "\n")

        # Write one item per line
        for item in items:
            values = []
            for header in headers:
                values.append(item.get(header, ""))
            f.write(",".join(values) + "\n")



# Adapted from: https://github.com/PRBonn/bonnetal/blob/master/train/common/trtCalibINT8.py
import numpy as np
import cv2
import os

import tensorrt as trt
import pycuda.autoinit
import pycuda.driver as cuda

TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE)


class EntropyCalibrator(trt.IInt8EntropyCalibrator2):
    """
    Prepare callibartion dataset for int8 callibration.

    :param model_path: Path to the model
    :type model_path: str
    :param calibration_files: callibaration image paths
    :type calibration_files: list
    :param batch_size: batch size
    :type batch_size: int
    :param h: input size height
    :type h: int
    :param w: input size width
    :type w: int
    :param means: means
    :type means: float
    :param stds: standard deviation
    :type stds: float

    """

    def __init__(self, model_path, calibration_files, batch_size, h, w, means, stds):
        """
        :param model_path: directory path where model will be loaded from
        :type model_path: str
        :param calibration_files: path to calibration image files
        :type calibration_files: str
        :param batch_size: batch_size
        :type batch_size: int
        :param h: height
        :type h: int
        :param w: width
        :type w: int
        :param means: means
        :type means: np.array
        :param stds: standard deviation
        :type stds: np.array
        """
        trt.IInt8EntropyCalibrator2.__init__(self)
        self.batch_size = batch_size
        self.h = h
        self.w = w
        self.input_size = [h, w]
        self.means = np.array(means, dtype=np.float32)
        self.stds = np.array(stds, dtype=np.float32)
        assert isinstance(calibration_files, list)
        self.calib_image_paths = calibration_files
        assert os.path.exists(model_path)
        self.cache_file = os.path.join(model_path, "model.trt.int8calib")
        self.shape = [self.batch_size, 3] + self.input_size
        self.device_input = cuda.mem_alloc(
            trt.volume(self.shape) * trt.float32.itemsize
        )
        self.indices = np.arange(len(self.calib_image_paths))
        np.random.shuffle(self.indices)

        def load_batches():
            """
            loads batches
            """
            for i in range(
                0, len(self.calib_image_paths) - self.batch_size + 1, self.batch_size
            ):
                indxes = self.indices[i : i + self.batch_size]
                paths = [self.calib_image_paths[i] for i in indxes]
                files = self.read_batch_file(paths)
                yield files

        self.batches = load_batches()

    def read_batch_file(self, filenames):
        """
        reads batch files

        :param filenames: list of file names
        :type filenames: List
        :returns: contiguous array
        :returntype: np.ascontiguousarray
        """
        tensors = []
        for filename in filenames:
            assert os.path.exists(filename)
            bgr_img = cv2.imread(filename)
            assert bgr_img.data
            bgr_img = cv2.resize(
                bgr_img, (self.w, self.h), interpolation=cv2.INTER_LINEAR
            )
            rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB).astype(np.float32)
            rgb_tensor = (rgb_img / 255.0 - self.means) / self.stds
            rgb_tensor = np.transpose(rgb_tensor, (2, 0, 1))
            rgb_tensor = np.ascontiguousarray(rgb_tensor, dtype=np.float32)
            tensors.append(rgb_tensor)
        return np.ascontiguousarray(tensors, dtype=np.float32)

    def get_batch_size(self):
        """
        returns batch size

        :returns: batch size
        :returntype: int
        """
        return self.batch_size

    # def get_batch(self, bindings, names):
    # https://docs.nvidia.com/deeplearning/tensorrt/api/python_api/infer/Int8/EntropyCalibrator2.html#tensorrt.IInt8EntropyCalibrator2.get_batch
    def get_batch(self, names):
        """
        returns a batch

        :param names:

        :returns: List[batch from self.batches iter]
        :returntype: List
        """
        try:
            data = next(self.batches)
            cuda.memcpy_htod(self.device_input, data)
            return [int(self.device_input)]
        except StopIteration:
            return None

    def read_calibration_cache(self):
        """
        reads calibration cache

        :returns: file content from self.cache_file
        :returntype: file.read()
        """
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                return f.read()

    def write_calibration_cache(self, cache):
        """
        writes cache to self.cache_file

        :param cache: cache to write
        :type cache: str
        """
        with open(self.cache_file, "wb") as f:
            f.write(cache)



import gdal
import numpy as np


def convert_pixels_to_geocoordinates(source_tiff_path, list_of_coords):

    """
    This function convert pixel coordinates into geo coordinates according to the transform informaton inside the source tiff file

    :param source_tiff_path: The source tiff file path from which the bounding box was predicted. Tiff file should contain geo-transform information
    :type source_tiff_path: string
    :param list_of_coords: list of coordinates in the form of ((x1, y1), (x2,y2), (x3,y3)...(xn, yn))
    :type bbox: List of list

    :returns: List of geo-coordinates

    """
    assert isinstance(source_tiff_path, str), "input source path should be string"

    tile = gdal.open(source_tiff_path)
    geotrans = tile.GetGeoTransform()

    # extract coefficients used to transform pixels to geo-grid points
    x_min = geotrans[0]
    x_size = geotrans[1]
    y_min = geotrans[3]
    y_size = geotrans[5]

    # create empty nx2 array with same length as origninal mask
    geocoords = np.empty([len(list_of_coords), 2], dtype=float)

    # convert all points to geo coordinates
    for i in range(len(list_of_coords)):
        geocoords[i][0] = list_of_coords[i][0] * x_size + x_min  # x pixel
        geocoords[i][1] = list_of_coords[i][1] * y_size + y_min  # y pixel

    return geocoords


def create_contours(coords):
    """
    Converting the coordinates of the masks to numpy array

    :param coords: List of Coordinates in the form the form of ((x1, y1), (x2,y2), (x3,y3)...(xn, yn))
    :type coords: List of lists

    :returns: list of contours in numpy format
    """
    cnts = []
    contours = coords
    maxlength = max(map(len, contours))

    for contour in contours:
        contour_arr = np.zeros(maxlength, dtype=np.int32)
        contour_arr[: len(contour)] = contour
        cnts.append(contour_arr)

    final_cnts = np.array(cnts)
    return final_cnts


def zoom_to_fill(image, mask, padding_val=0):
    """
    Use the mask of the object to center the object in the image and zoom to that object

    :param image: Input Image
    :type image: numpy array
    :param mask: Mask of the object that you want to center
    :type mask: numpy array
    :param padding_val: number of pixel padding to add around the centered object
    :type padding_val: int

    :returns: Numpy image with centered object

    """
    (y, x) = np.where(mask == 255)
    (topy, topx) = (np.min(y), np.min(x))
    (bottomy, bottomx) = (np.max(y), np.max(x))
    image = image[
        topy - padding_val : bottomy + padding_val,
        topx - padding_val : bottomx + padding_val,
    ]

    return image


def add_padding(image, padding_height, padding_width, color=(0, 0, 0)):
    """
    add padding around the image so that buildings touching the corners can be bought to center
    create new image of desired size and color (black) for padding
    This is necessary as corner buildings will not be detected by countour detection algorithm if the building edges are outsite of the image

    :param images: Image that you want padding around
    :type image: Numpy array
    :param padding_height: Amount of padding in y direction
    :type padding_height: int
    :param padding_width: Amount of padding in x direction
    :param padding_width: int
    :param color: RGB color of the padding
    :type color: 1x3 tuple

    :returns: Image with padding
    """
    ht, wd, cc = image.shape
    ww = wd + padding_width
    hh = ht + padding_height
    padded = np.full((hh, ww, cc), color, dtype=np.uint8)

    # compute center offset
    xx = (ww - wd) // 2
    yy = (hh - ht) // 2

    # copy img image into center of RGB_padded image
    padded[yy : yy + ht, xx : xx + wd] = image

    return padded



import os
import cv2
import numpy as np
import torch
import onnxruntime

from ..utils import Params

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class ValidateOnnx:
    """Validate the prediction of onnx model with the pytorch model

    :param model_dir: Path to folder containing `params.json` file and `model.onnx` in `models` directory.
    :type model_dir: str

    """

    def __init__(self, model_dir: str):
        """Load the model with path corresponding to `save_model_path` in `param.json` file."""
        json_path = os.path.join(model_dir, "params.json")
        assert os.path.isfile(
            json_path
        ), "No json configuration file found at {}".format(json_path)
        params = Params(json_path)
        self.model_path = os.path.join(model_dir, "models/")
        self.onnx_path = os.path.join(self.model_path, "model.onnx")

        self.pytorch_path = params.save_model_path
        self.data_h, self.data_w, self.data_d = (
            params.height,
            params.width,
            params.input_channels,
        )
        means, stds = params.means, params.stds
        self.means = np.array(means, dtype=np.float32)
        self.stds = np.array(stds, dtype=np.float32)

    @staticmethod
    def _to_numpy(tensor):
        return (
            tensor.detach().cpu().numpy()
            if tensor.requires_grad
            else tensor.cpu().numpy()
        )

    def calibrate(self, img_paths: list, tol: float, root_dir: str = None):
        """Perform calibration for pytorch and onnx model for given relative threshold

        :param img_paths: List of some sample image paths to calibrate the prediction
        :type img_paths: list
        :param tol: Relative tolerance to compare the onnx and pytorch predictions
        :type tol: float
        :param root_dir: Root directory to load images if the paths are not absolute
        :type root_dir: str
        :returns: Prints the assertion output if there's mismatch in prediction of pytorch and onnx model for given threshold

        """
        ort_session = onnxruntime.InferenceSession(self.onnx_path)
        model = torch.load(self.pytorch_path).to(device)
        model.eval()
        for img_path in img_paths:
            if root_dir is not None:
                bgr_img = cv2.imread(os.path.join(root_dir, img_path))
            else:
                bgr_img = cv2.imread(img_path)
            bgr_img = cv2.resize(
                bgr_img, (self.data_w, self.data_h), interpolation=cv2.INTER_LINEAR
            )
            # check if network is RGB or mono
            if self.data_d == 3:
                # get make rgb
                rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB)
            elif self.data_d == 1:
                # get grayscale
                rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2GRAY)
            else:
                raise NotImplementedError(
                    "Network has to have 1 or 3 channels. Anything else must be implemented."
                )
            # to tensor
            rgb_tensor = torch.from_numpy(rgb_img)
            # permute and normalize
            rgb_tensor = (rgb_tensor.float() / 255.0 - self.means) / self.stds
            rgb_tensor = rgb_tensor.permute(2, 0, 1)
            # add batch dimension
            rgb_tensor = rgb_tensor.unsqueeze(0)

            # pytorch results
            torch_out = model(rgb_tensor.to(device))

            # compute ONNX Runtime output prediction
            ort_inputs = {ort_session.get_inputs()[0].name: self._to_numpy(rgb_tensor)}
            ort_outs = ort_session.run(None, ort_inputs)
            try:
                # compare ONNX Runtime and PyTorch results
                np.testing.assert_allclose(
                    self._to_numpy(torch_out), ort_outs[0], rtol=tol, atol=1e-05
                )
                print(f"Passed {img_path} with relative tol={tol}")
                print("-" * 15)
            except AssertionError as e:
                print(f"Failed {img_path} with relative tol={tol}")
                print(e)
                print("-" * 30)



from simpletransformers.classification import MultiLabelClassificationModel
from simpletransformers.classification import ClassificationModel, ClassificationArgs
import lightgbm as lgb
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import VotingClassifier
import xgboost as xgb
from catboost import CatBoostClassifier
import pandas as pd
import numpy as np
import sklearn
from typing import List
from pprint import pprint


def multilabel_classification(
    model_type, model_name, model_params, train_params, eval_params, train_set, eval_set
):
    """
    Train and evaluate a multilabel classification model.

    :param model_type: One of the model types from the supported models, https://simpletransformers.ai/docs/classification-specifics/#supported-model-types
    :param model_type: str
    :param model_name: Name of the model to use. Find it on the official simpletransformer documentation, https://huggingface.co/transformers/pretrained_models.html
    :type model_name: str
    :param model_params: Model parameters.
    :type model_params: dict
    :param train_params: Training parameters.
    :type train_params: dict
    :param eval_params: Evaluation parameters.
    :type eval_params: dict
    :param train_set: Training set.
    :type train_set: pandas.DataFrame
    :param eval_set: Evaluation set.
    :type eval_set: pandas.DataFrame
    :returns:
        model: fine-tuned simpletransformers model
        results: dict: ex: {"LRAP": 0.98, "eval_loss": 0.15}
    """

    assert type(train_df) == pd.DataFrame, "train_df must be a pandas.DataFrame"
    assert type(eval_df) == pd.DataFrame, "eval_df must be a pandas.DataFrame"

    model = MultiLabelClassificationModel(
        model_type, model_name, use_cuda=False, num_labels=6
    )

    model.train_model(train_df)
    result, model_outputs, wrong_predictions = model.eval_model(eval_df)

    return result, model_outputs, wrong_predictions


def voting_classifier(
    features_train: np.array,
    labels_train,
    features_test: np.array,
    lgbm_params: List = None,
    rf_params: List = None,
    xgb_params: List = None,
    catboost_params: List = None,
) -> sklearn.VotingClassifier:
    """
    Train a voting classifier.

    :param features_train: Training features.
    :type features_train: numpy.ndarray
    :param labels_train: Training labels.
    :type labels_train: List
    :param features_test: Testing features.
    :type features_test: numpy.ndarray
    :param lgbm_params: Parameters for lightgbm.
    :type lgbm_params: dict
    :param rf_params: Parameters for sklearn.ensemble.RandomForestClassifier.
    :type rf_params: dict
    :param xgb_params: Parameters for xgboost.
    :type xgb_params: dict
    :param catboost_params: Parameters for catboost.
    :type catboost_params: dict
    :returns:
        model: fine-tuned voting classifier model
    """

    lgbm_params = lgbm_params if lgbm_params else {}
    rf_params = rf_params if rf_params else {}
    xgb_params = xgb_params if xgb_params else {}
    catboost_params = catboost_params if catboost_params else {}

    lgbm_model = lgb.LGBMClassifier(**lgbm_params)
    rf_model = RandomForestClassifier(**rf_params)
    xgb_model = xgb.XGBClassifier(**xgb_params)
    catboost_model = CatBoostClassifier(**catboost_params)

    voting_model = VotingClassifier(
        estimators=[
            ("lgbm", lgbm_model),
            ("rf", rf_model),
            ("xgb", xgb_model),
            ("catboost", catboost_model),
        ],
        voting="soft",
    )

    vot.fit(features_train, labels_train)
    pprint(vot)

    return voting_model


def multiclass_classification(
    model_type, model_name, epochs, train_set, train_labels, eval_set, eval_labels
):
    """
    Train and evaluate a multilabel classification model.
    :param model_type: One of the model types from the supported models, https://simpletransformers.ai/docs/classification-specifics/#supported-model-types
    :param model_type: str
    :param model_name: Name of the model to use. Find it on the official simpletransformer documentation, https://huggingface.co/transformers/pretrained_models.html
    :type model_name: str
    :param epochs: Number of epochs to train.
    :type epochs: int
    :param train_set: Training set.
    :type train_set: List[str]
    :param train_labels: Training labels.
    :type train_labels: List[int]
    :param eval_set: Evaluation set.
    :type eval_set: List[str]
    :param eval_labels: Evaluation labels.
    :type eval_labels: List[int]
    :returns:
        model: fine-tuned simpletransformers model
        results: dict: ex: {"mcc": 0.98, "eval_loss": 0.15}
        mcc is Matthews correlation coefficient (https://scikit-learn.org/stable/modules/generated/sklearn.metrics.matthews_corrcoef.html)
    """

    assert type(train_set) == list, "train_set must be a list"
    assert type(train_labels) == list, "train_labels must be a list"
    assert type(eval_set) == list, "eval_set must be a list"
    assert type(eval_labels) == list, "eval_labels must be a list"

    epochs = 1 if epochs == 0 else epochs

    train_df = pd.DataFrame({"text": train_set, "label": train_labels})

    eval_df = pd.DataFrame({"text": eval_set, "label": eval_labels})

    model_args = ClassificationArgs(num_train_epochs=1)

    model = ClassificationModel(
        model_type, model_name, num_labels=len(train_df.label.unique()), args=model_args
    )

    model.train_model(train_df)

    result, _, _ = model.eval_model(eval_df)

    return model, result
